"""Physical constants

"""
import scipy.constants
import scipy.special


c = scipy.constants.c
G = scipy.constants.G
g = scipy.constants.g
E0 = scipy.constants.epsilon_0
hbar = scipy.constants.hbar
kB = scipy.constants.k
yr = scipy.constants.year
AU = scipy.constants.astronomical_unit
parsec = scipy.constants.parsec
Mpc = scipy.constants.parsec * 1e6

# FIXME: use astropy for the following astrophysical/cosmological
# constants
R_earth = 6.3781e6
SolarMassParameter = 1.3271244e20
MSol = 1.3271244e20 / scipy.constants.G
# http://physics.nist.gov/cuu/Constants/
H0 = 67110
# http://arxiv.org/pdf/1303.5076v3.pdf
omegaM = 0.3175,
omegaLambda = 1 - 0.3175

# bessel function zeros
BESSEL_ZEROS = scipy.special.jn_zeros(1, 300)
J0M = scipy.special.jn(0, BESSEL_ZEROS)



import itertools
import collections
import numpy as np
import scipy.interpolate

from . import logger
from .trace import BudgetTrace


def precomp(*precomp_funcs, **precomp_fmaps):
    """BudgetItem.calc decorator to add pre-computed functions

    This is intended to decorate BudgetItem.calc() methods with common
    functions whose return value are cached for later use.  The
    functions are supplied with the `freq` array and `ifo` Struct
    attributes as arguments, and their return values are provided as
    keyword arguments to calc().

    For example, if a calc method is defined as:

      @precomp(foo=precomp_foo)
      @precomp(bar=precomp_bar)
      def calc(self, foo, bar):
          ...

    then when the budget is run the precomp functions will be executed
    before calc(), roughly the equivalent of:

      foo = precomp_foo(self.freq, self.ifo)
      bar = precomp_bar(self.freq, self.ifo)
      psd = calc(foo=foo, bar=bar)

    The execution state of each precomp function is cached so that the
    same function is not needlessly executed multiple times.

    """
    def decorator(func):
        if precomp_funcs:
            try:
                func._precomp_list.extend(precomp_funcs)
            except AttributeError:
                func._precomp_list = list(precomp_funcs)

        if precomp_fmaps:
            try:
                func._precomp_mapped.update(precomp_fmaps)
            except AttributeError:
                func._precomp_mapped = dict(precomp_fmaps)
        return func
    return decorator


def _precomp_recurse_mapping(func, freq, ifo, _precomp):
    """Recursively execute @precomp decorator functions

    Recurses down functions which may themselves have precomp
    decorators, building a **kwargs mapping to pass to the wrapped
    function call.

    """
    for pc_func in itertools.chain(
            getattr(func, '_precomp_list', []),
            getattr(func, '_precomp_mapped', {}).values(),
    ):
        if pc_func in _precomp:
            continue
        pc_map = _precomp_recurse_mapping(pc_func, freq, ifo, _precomp=_precomp)
        logger.debug("precomp {}".format(pc_func))
        _precomp[pc_func] = pc_func(freq, ifo, **pc_map)

    kwargs = {
        name: _precomp[pc_func] for name, pc_func in getattr(func, '_precomp_mapped', {}).items()
    }
    return kwargs


def quadsum(data):
    """Calculate quadrature sum of list of data arrays.

    Provided data are assumed to be power-referred, so this is a
    simple point-by-point sum.

    NaNs in sum elements do not contribute to sum.

    """
    return np.nansum(data, 0)


class BudgetItem:
    """GWINC BudgetItem class

    """
    def load(self):
        """Overload method for initial loading of static data.

        """
        return None

    def update(self, **kwargs):
        """Overload method for updating data.

        By default any keyword arguments provided are written directly
        as attribute variables (as with __init__).

        When overloading this method it is recommended to execute the
        method from the base class with e.g.:

          super().update(**kwargs)

        """
        for key, val in kwargs.items():
            setattr(self, key, val)

    def calc(self):
        """Overload method for final PSD calculation.

        Should return an array of power-referenced values evaluated at
        all evaluation frequencies (self.freq).

        """
        return None

    def _calc(self, _precomp=None):
        """internal function to call calc with precomp evaluation"""
        pcmap = _precomp_recurse_mapping(self.calc, self.freq, self.ifo, _precomp)
        logger.debug("calc {}".format(self.name))
        return self.calc(**pcmap)

    ##########

    def __init__(self, freq=None, **kwargs):
        """Initialize budget item.

        The primary argument should be the evaluation frequency array.
        If not provided, then a pre-defined `freq` attribute of the
        BudgetItem class should exist.

        Additional keyword arguments are written as attribute
        variables to the initialized object.

        """
        if freq is not None:
            assert isinstance(freq, np.ndarray)
            self.freq = freq
        self.ifo = None
        for key, val in kwargs.items():
            setattr(self, key, val)
        self._loaded = False
        self._precomp = dict()

    @property
    def name(self):
        """"Name of this BudgetItem class."""
        return self.__class__.__name__

    def __str__(self):
        # FIXME: provide info on internal state (load/update/calc/etc.)
        return '<{} {}>'.format(
            ', '.join([c.__name__ for c in self.__class__.__bases__]),
            self.name,
        )

    def interpolate(self, freq, data):
        """Interpolate data to the evaluation frequencies.

        """
        func = scipy.interpolate.interp1d(
            freq, data,
            kind='nearest',
            copy=False,
            assume_sorted=True,
            bounds_error=False,
            fill_value=np.nan,
        )
        return func(self.freq)


class Calibration(BudgetItem):
    """GWINC Calibration class

    BudgetItem that represents a calibration transfer function for a
    Noise.  The calc() method should return a transfer function
    amplitude array evaluated at the evaluation frequencies supplied
    at initialization and available in the `freq` array attribute
    (self.freq).

    """
    def __call__(self, data):
        """Calibrate input data.

        Returns calibrated version of input data array,
        e.g. point-by-point product of data and calibration arrays.

        """
        cal = self._calc()
        assert data.shape == cal.shape, \
            "data shape does not match calibration ({} != {})".format(data.shape, cal.shape)
        return data * cal


class Noise(BudgetItem):
    """GWINC Noise class

    BudgetItem that represents a PSD noise calculation.  The calc()
    method should return the noise PSD spectrum array evaluated at the
    evaluation frequencies supplied at initialization and available in
    the `freq` array attribute (self.freq).

    """

    style = {}
    """Trace plot style dictionary"""

    def _make_trace(self, psd=None, budget=None):
        return BudgetTrace(
            name=self.name,
            style=self.style,
            freq=self.freq,
            psd=psd,
            budget=budget,
        )

    def calc_trace(self, calibration=1, calc=True, _precomp=None):
        """Calculate noise and return BudgetTrace object

        `calibration` should either be a scalar or a len(self.freq)
        array that will be multiplied to the output PSD of the budget
        and all sub noises.

        If `calc` is False, the noise will not be calculated and the
        trace PSD will be None.  This is useful for just getting the
        trace style info.

        """
        if _precomp is None:
            _precomp = dict()

        total = None
        if calc:
            total = self._calc(_precomp) * calibration

        return self._make_trace(psd=total)

    def run(self, **kwargs):
        """Run full budget and return BudgetTrace object.

        Roughly the equivalent running load(), update(), and
        calc_trace() in sequence.  Keyword arguments are passed to the
        update() method.

        NOTE: The load status is cached such that subsequent calls to
        this method will not re-execute the load() method.

        NOTE: The update() method is only run if keyword arguments
        (`kwargs`) are supplied, or if the `ifo` attribute has
        changed.

        """
        if not self._loaded:
            self.load()
            self._loaded = True

        ifo = kwargs.get('ifo', getattr(self, 'ifo'))
        if ifo:
            if not hasattr(ifo, '_orig_keys'):
                logger.debug("new ifo detected")
                ifo._orig_keys = tuple(k for k, v in ifo.walk())
                ifo_hash = ifo.hash()
                kwargs['ifo'] = ifo
            else:
                ifo_hash = ifo.hash(ifo._orig_keys)
                if ifo_hash != getattr(self, '_ifo_hash', 0):
                    logger.debug("ifo hash change")
                    kwargs['ifo'] = self.ifo
            self._ifo_hash = ifo_hash

        if kwargs:
            self.update(**kwargs)
            # clear precomp cache
            self._precomp = dict()

        return self.calc_trace(_precomp=self._precomp)


class Budget(Noise):
    """GWINC Budget class

    This is a Noise that represents a budget of multiple sub noises.

    The `noises` attribute of this class should list constituent Noise
    classes.  Each element can be either a single Noise class, or a
    tuple of (Noise, Calibration) classes, e.g.:

    noises = [
        Thermal,
        (Shot, Sensing),
    ]

    When this object is initialized, all sub noises and calibrations
    are initialized.  Pre-defined load() and update() methods call the
    load() and update() methods of all sub noises and calibrations.
    When calc() is called, the PSD is calculated for all sub noises,
    the relevant calibration is evaluated and applied, and the
    quadrature sum of all calibrated consituent noises is returned.

    Additionally, a `calibrations` attribute may define a list of
    common calibrations to apply to all noises, e.g.:

    calibrations = [
        Strain,
    ]

    Finally, a `references` attribute may be defined, similar to the
    `noises` attribute described above except that the specified
    noises do not contribute to the overall budget total, e.g.:

    references = [
        strain_data_20200120,
    ]

    NOTE: if an `ifo` attribute is defined it is always passed as an
    initialization argument to sub noises.

    """

    noises = []
    """List of constituent noise classes, or (noise, cal) tuples"""

    calibrations = []
    """List of calibrations to be applied to all budget noises (not references)"""

    references = []
    """List of reference noise classes, or (ref, cal) tuples"""

    accumulate = quadsum
    """Method used to calculate total noise trace"""

    def __init__(self, freq=None, noises=None, **kwargs):
        """Initialize Budget object.

        See BudgetItem for base initialization arguments.

        If a `noises` keyword argument is provided it should be an
        iterable of noise names (constituent or reference) which will
        be used to filter the noises initialized in this budget.

        """
        super().__init__(freq, **kwargs)
        # store kwargs for later use
        self.kwargs = kwargs
        # record the frequency array as a kwarg if it's definied as a
        # class attribute
        if freq is not None:
            self.kwargs['freq'] = freq
        else:
            self.kwargs['freq'] = getattr(self, 'freq', None)
        # FIXME: special casing the ifo kwarg here, in case it's
        # defined as a class attribute rather than passed at
        # initialization.  we do this because we're not defining a
        # standard way to extract IFO variables that get passed around
        # in a reasonable way.  how can we clarify this?
        if 'ifo' not in kwargs and getattr(self, 'ifo', None):
            self.kwargs['ifo'] = getattr(self, 'ifo', None)
        # all noise objects keyed by name
        self._noise_objs = collections.OrderedDict()
        # all cal objects keyed by name
        self._cal_objs = {}
        # set of calibration names to apply to noise
        self._noise_cals = collections.defaultdict(set)
        # set of all constituent budget noise names
        self._budget_noises = set()
        # initialize all noise objects
        for nc in self.noises:
            name = self.__init_noise(nc, noises)
            if name:
                self._budget_noises.add(name)
        # initialize common calibrations and add to all budget noises
        for cal in self.calibrations:
            self.__add_calibration(cal, self._budget_noises)
        # initialize references, without common calibrations
        for nc in self.references:
            self.__init_noise(nc, noises)
        # error if requested noise is not present
        if noises:
            sset = set(noises)
            nset = set([name for name in self._noise_objs.keys()])
            if not sset <= nset:
                raise AttributeError("unknown noise terms: {}".format(' '.join(sset-nset)))

    def __init_noise(self, nc, noise_filt):
        cal = None
        if isinstance(nc, (list, tuple)):
            noise = nc[0]
            cals = nc[1:]
        else:
            noise = nc
            cals = []
        noise_obj = noise(
            **self.kwargs
        )
        name = noise_obj.name
        if noise_filt and name not in noise_filt:
            return
        logger.debug("init {}".format(noise_obj))
        self._noise_objs[name] = noise_obj
        for cal in cals:
            self.__add_calibration(cal, [name])
        return name

    def __add_calibration(self, cal, noises):
        cal_obj = cal(
            **self.kwargs
        )
        name = cal_obj.name
        if name not in self._cal_objs:
            logger.debug("init {}".format(cal_obj))
            self._cal_objs[name] = cal_obj
        # register noises for this calibration
        for noise in noises:
            self._noise_cals[noise].add(name)
        return name

    def __getitem__(self, name):
        """Get a (possibly nested) sub-BudgetItem.

        """
        try:
            name, rest = name.split('.', 1)
        except ValueError:
            rest = None
        try:
            o = self._noise_objs[name]
        except KeyError:
            try:
                o = self._cal_objs[name]
            except KeyError:
                raise KeyError("unknown noise or cal name '{}".format(name))
        if rest:
            return o[rest]
        else:
            return o

    def get(self, key, default=None):
        """Get a (possibly nested) sub-BudgetItem.

        """
        try:
            return self[key]
        except KeyError:
            return default

    def keys(self):
        """Iterate over budget noise names."""
        return iter(self._noise_objs.keys())

    def values(self):
        """Iterate over budget noise objects."""
        return iter(self._noise_objs.values())

    def items(self):
        """Iterate over budget noise (name, object) tuples."""
        return iter(self._noise_objs.items())

    def __iter__(self):
        return iter(self.keys())

    def walk(self):
        """Walk recursively through every BudgetItem in the budget.

        This includes Noise, Calibration and Budget objects, as well
        as any decendents of Budget objects.

        For each leaf item yields a tuple of all ancestor objects,
        e.g.:

          (self)
          (self, BudgetItem)
          (self, ChildBudget1)
          (self, ChildBudget1, BudgetItem)
          ...

        """
        yield (self,)
        for item in itertools.chain(
                self._cal_objs.values(),
                self._noise_objs.values()):
            if isinstance(item, Budget):
                for i in item.walk():
                    yield (self,) + i
            else:
                yield (self, item)

    def load(self):
        """Load all noise and cal objects."""
        for name, item in itertools.chain(
                self._cal_objs.items(),
                self._noise_objs.items()):
            logger.debug("load {}".format(item))
            item.load()

    def update(self, _precomp=None, **kwargs):
        """Recursively update all noise and cal objects with supplied kwargs.

        See BudgetItem.update() for more info.

        """
        for key, val in kwargs.items():
            setattr(self, key, val)

        for name, item in itertools.chain(
                self._cal_objs.items(),
                self._noise_objs.items()):
            logger.debug("update {}".format(item))
            item.update(**kwargs)

    def calc_noise(self, name, calibration=1, calc=True, _cals=None, _precomp=None):
        """Return calibrated individual noise BudgetTrace.

        The noise and calibration transfer functions are calculated,
        and the calibrated noise BudgetTrace is returned.
        `calibration` is an overall calculated calibration to apply to
        the noise.

        """
        if _precomp is None:
            _precomp = dict()
        # must copy first if using mul-in-place or we clobber other cals
        calibration = np.ones_like(self.freq) * calibration
        for cal in self._noise_cals[name]:
            if _cals:
                calibration *= _cals[cal]
            else:
                obj = self._cal_objs[cal]
                calibration *= obj._calc(_precomp)
        noise = self._noise_objs[name]
        return noise.calc_trace(
            calibration=calibration,
            calc=calc,
            _precomp=_precomp,
        )

    def calc_trace(self, calibration=1, calc=True, _precomp=None):
        """Calculate all budget noises and return BudgetTrace object

        `calibration` should either be a scalar or a len(self.freq)
        array that will be multiplied to the output PSD of the budget
        and all sub noises.

        If `calc` is False, the noise will not be calculated and the
        trace PSD will be None.  This is useful for just getting the
        trace style info.

        """
        if _precomp is None:
            _precomp = dict()

        _cals = {}
        if calc:
            for name, cal in self._cal_objs.items():
                _cals[name] = cal._calc(_precomp)
        budget = []
        for name in self._noise_objs:
            trace = self.calc_noise(
                name,
                calibration=calibration,
                calc=calc,
                _cals=_cals,
                _precomp=_precomp,
            )
            budget.append(trace)
        total = type(self).accumulate([trace.psd for trace in budget if trace.name in self._budget_noises])
        return self._make_trace(
            psd=total, budget=budget
        )



'''Functions to calculate quantum noise

'''
from __future__ import division
import numpy as np
from numpy import pi, sqrt, arctan, sin, cos, exp, log10, conj
from copy import deepcopy
from collections.abc import Sequence

from .. import logger
from .. import const
from ..struct import Struct


def sqzOptimalSqueezeAngle(Mifo, eta):
    vHD = np.array([[sin(eta), cos(eta)]])
    H = getProdTF(vHD, Mifo)[0]
    alpha = arctan(abs(H[1]), abs(H[0]))
    return alpha


def getSqzParams(ifo):
    """Determine squeezer type, if any, and extract common parameters

    Returns a struct with the following attributes:
    SQZ_DB: squeezing in dB
    ANTISQZ_DB: anti-squeezing in dB
    alpha: freq Indep Squeeze angle [rad]
    lambda_in: loss to squeezing before injection [Power]
    etaRMS: quadrature noise [rad]
    eta: homodyne angle [rad]
    """
    params = Struct()

    if 'Squeezer' not in ifo:
        sqzType = None
    else:
        sqzType = ifo.Squeezer.get('Type', 'Freq Independent')

    params.sqzType = sqzType

    # extract squeezer parameters
    if sqzType is None:
        params.SQZ_DB = 0
        params.ANTISQZ_DB = 0
        params.alpha = 0
        params.lambda_in = 0
        params.etaRMS = 0

    else:
        params.SQZ_DB = ifo.Squeezer.AmplitudedB
        params.ANTISQZ_DB = ifo.Squeezer.get('AntiAmplitudedB', params.SQZ_DB)
        params.alpha = ifo.Squeezer.SQZAngle
        params.lambda_in = ifo.Squeezer.InjectionLoss
        params.etaRMS = ifo.Squeezer.get('LOAngleRMS', 0)

    # Homodyne Readout phase
    eta_orig = ifo.Optics.get('Quadrature', Struct()).get('dc', None)

    ifoRead = ifo.get('Squeezer', Struct()).get('Readout', None)
    if ifoRead is None:
        eta = eta_orig
        if eta_orig is None:
            raise Exception("must add Quadrature.dc or Readout...")

    elif ifoRead.Type == 'DC':
        eta = np.sign(ifoRead.fringe_side) * \
            np.arccos((ifoRead.defect_PWR_W / ifoRead.readout_PWR_W)**.5)

    elif ifoRead.Type == 'Homodyne':
        eta = ifoRead.Angle

    else:
        raise Exception("Unknown Readout Type")

    if eta_orig is not None:
        # logger.warn((
        #     'Quadrature.dc is redundant with '
        #     'Squeezer.Readout and is deprecated.'
        # ))
        if eta_orig != eta:
            raise Exception("Quadrature.dc inconsistent with Readout eta")

    params.eta = eta

    return params


######################################################################
# Main quantum noise function
######################################################################


def shotrad(f, ifo, sustf, power):
    """Quantum noise strain spectrum

    :f: frequency array in Hz
    :ifo: gwinc IFO Struct
    :sustf: suspension transfer function struct
    :power: gwinc power Struct

    :returns: displacement noise power spectrum at :f:

    corresponding author: mevans

    """
    ######################################################################
    # Vacuum sources will be individually tracked with the Mnoise dict   #
    # The keys are the following                                         #
    #   * arm: arm cavity losses                                         #
    #   * SEC: SEC losses                                                #
    #   * ASvac: AS port vacuum                                          #
    #   * pd: readout losses                                             #
    #   * injection: injection losses                                    #
    #   * phase: quadrature phase noise                                  #
    #                                                                    #
    # If there is a filter cavity there are an additional set of keys    #
    # starting with 'FC' for the filter cavity losses                    #
    #                                                                    #
    # If there is an output filter cavity the key 'OFC' gives its losses #
    ######################################################################

    # call the main IFO Quantum Model
    if 'Type' not in ifo.Optics or ifo.Optics.Type == 'SignalRecycled':
        coeff, Mifo, Msig, Mn = shotradSignalRecycled(f, ifo, sustf, power)
    elif ifo.Optics.Type == 'SignalRecycledBnC':
        coeff, Mifo, Msig, Mn = shotradSignalRecycledBnC(f, ifo, power)
    else:
        raise ValueError('Unrecognized IFO type ' + ifo.Optics.Type)

    # separate arm and SEC loss
    Mnoise = dict(arm=Mn[:, :2, :], SEC=Mn[:, 2:, :])

    sqz_params = getSqzParams(ifo)

    if sqz_params.sqzType == 'Optimal':
        # compute optimal squeezing angle
        sqz_params.alpha = sqzOptimalSqueezeAngle(Mifo, sqz_params.eta)

    vHD = np.array([[sin(sqz_params.eta), cos(sqz_params.eta)]])

    def homodyne(signal):
        """Readout the eta quadrature of the signal signal
        """
        return np.squeeze(np.sum(abs(getProdTF(vHD, signal))**2, axis=1))

    # optomechanical plant
    lambda_PD = 1 - ifo.Optics.PhotoDetectorEfficiency  # PD losses
    Msig = Msig * sqrt(1 - lambda_PD)
    plant = homodyne(Msig)

    # get all the losses in the squeezed quadrature
    Mnoise = propagate_noise_fc_ifo(
        f, ifo, Mifo, Mnoise, sqz_params, sqz_params.alpha)

    # add quadrature phase fluctuations if any
    if sqz_params.etaRMS:
        # get all losses in the anti-squeezed quadrature
        Mnoise_antisqz = propagate_noise_fc_ifo(
            f, ifo, Mifo, Mnoise, sqz_params, sqz_params.alpha + pi/2)

        # sum over these since they're not individually tracked
        var_antisqz = np.zeros_like(f)
        for nn in Mnoise_antisqz.values():
            var_antisqz += homodyne(nn)

        # The total variance is
        # V(alpha) * cos(etaRMS)^2 + V(alpha + pi/2) * sin(etaRMS)^2
        Mnoise = {key: cos(sqz_params.etaRMS)**2 * homodyne(nn)
                  for key, nn in Mnoise.items()}
        Mnoise['phase'] = sin(sqz_params.etaRMS)**2 * var_antisqz

    else:
        Mnoise = {key: homodyne(nn) for key, nn in Mnoise.items()}

    # calibrate into displacement
    # coeff can be removed if shotradSignalRecycledBnC isn't used
    Mnoise = {key: coeff*nn/plant for key, nn in Mnoise.items()}
    psd = {key: nn * ifo.Infrastructure.Length**2 for key, nn in Mnoise.items()}

    return psd


######################################################################
# The following two compile functions generate analytic expressions
# that are hard coded in shotradSignalRecycled and are not called
# when shotrad is called
######################################################################


def compile_ARM_RES_TF():
    import sympy as sp
    ID = sp.eye(2)
    rITM, tArm, exp_2jOmegaL_c, K = sp.symbols('rITM tArm exp_2jOmegaL_c K')
    ARM = tArm * exp_2jOmegaL_c * sp.Matrix([[1, 0], [-K, 1]])
    ARM_RES = (ID - rITM*ARM)**-1

    subexprs, ARM_RES_expr = sp.cse(ARM_RES)
    for expr in subexprs:
        print(str(expr[0]), '=', str(expr[1]))
    print('RES', '=', str(ARM_RES_expr[0]).replace('Matrix', 'np.array').replace(', 0]', ', np.zeros(nf)]'))


def compile_SEC_RES_TF():
    import sympy as sp
    ID = sp.eye(2)
    phi, exp_1jOmegal_c, tArm, exp_2jOmegaL_c, K, r00, r10, r11, R, T, rITM, tSR, rho = sp.symbols('phi exp_1jOmegal_c tArm exp_2jOmegaL_c K r00 r10 r11 R T rITM tSR rho')
    SEr = sp.Matrix([[sp.cos(phi), sp.sin(phi)], [-sp.sin(phi), sp.cos(phi)]])
    SE = SEr * exp_1jOmegal_c
    ARM = tArm * exp_2jOmegaL_c * sp.Matrix([[1, 0], [-K, 1]])
    ARM_RES = sp.Matrix([[r00, 0], [r10, r11]])
    rho_ARM = ARM_RES * ((R + T) * ARM - rITM * ID)
    SEC = tSR * SE * rho_ARM * SE
    SEC_RES = (ID + rho*SEC)**-1

    subexprs, SEC_RES_expr = sp.cse(SEC_RES)
    for expr in subexprs:
        print(str(expr[0]), '=', str(expr[1]))
    print('RES', '=', str(SEC_RES_expr[0]).replace('Matrix', 'np.array'))


######################################################################
# Main IFO quantum models
######################################################################

def shotradSignalRecycled(f, ifo, sustf, power):
    """Quantum noise model for signal recycled IFO (see shotrad for more info)

    New version July 2016 by JH based on transfer function formalism

    coeff = frequency dependent overall noise coefficient (Nx1)
            (not required anymore, but kept for compatibility with shotrad.m)
    Mifo = IFO input-output relation for the AS port
    Msig = signal transfer to the AS port
    Mnoise = noise fields produced by losses in the IFO at the AS port

    """
    lambda_ = ifo.Laser.Wavelength               # Laser Wavelength [m]
    hbar = const.hbar                            # Plancks Constant [Js]
    c = const.c                                  # SOL [m/s]
    omega_0 = 2*pi*c/lambda_                     # Laser angular frequency [rads/s]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    L = ifo.Infrastructure.Length                # Length of arm cavities [m]
    l = ifo.Optics.SRM.CavityLength              # SRC Length [m]
    T = ifo.Optics.ITM.Transmittance             # ITM Transmittance [Power]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    bsloss = ifo.Optics.BSLoss                   # BS Loss [Power]
    mismatch = 1 - ifo.Optics.coupling           # Mismatch
    mismatch = mismatch + ifo.TCS.SRCloss        # Mismatch

    # BSloss + mismatch has been incorporated into a SRC Loss
    lambda_SR = 1 - (1 - mismatch) * (1 - bsloss)  # SR cavity loss [Power]

    tau = sqrt(ifo.Optics.SRM.Transmittance)     # SRM Transmittance [amplitude]
    rho = sqrt(1 - tau**2)                       # SRM Reflectivity [amplitude]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ds = ifo.Optics.SRM.Tunephase                # SRC Detunning
    phi = ds/2

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    lambda_arm = 1 - (1 - ifo.Optics.Loss)**2 * (1 - ifo.Optics.ETM.Transmittance)

    R = 1 - T - ifo.Optics.Loss                  # ITM Reflectivity [Power]

    P = power.parm                               # use precomputed value

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    nf = len(f)

    ID = np.array([[np.ones(nf), np.zeros(nf)], [np.zeros(nf), np.ones(nf)]])

    # transfer matrices for dark port input and signal field
    Mifo = np.zeros((2, 2, nf), dtype=complex)
    Msig = np.zeros((2, 1, nf), dtype=complex)

    # transfer matrices for SEC and arm loss fields
    Mp = np.zeros((2, 2, nf), dtype=complex)
    Mn = np.zeros((2, 2, nf), dtype=complex)

    # SRC rotation matrix
    SEr = np.array([[np.tile(cos(phi), nf), np.tile(sin(phi), nf)],
                    [np.tile(-sin(phi), nf), np.tile(cos(phi), nf)]])

    # some precomputed parameters
    tITM = sqrt(T)
    rITM = sqrt(R)

    tArm = sqrt(1 - lambda_arm)
    tSR = sqrt(1 - lambda_SR)
    tSig = sqrt((1 - lambda_arm / 2) * (1 - lambda_SR / 2))
    RT_SRM = rho**2 + tau**2

    lossArm = sqrt(lambda_arm)
    lossSR = sqrt(lambda_SR)

    Omega = 2*pi*f                                          # Signal angular frequency [rad/s]
    h_SQL = sqrt(8 * hbar * np.abs(sustf.tst_suscept)) / L  # SQL Strain
    K = - 16 * P * omega_0 / c**2 * sustf.tst_suscept

    # arm cavity
    exp_2jOmegaL_c = exp(2j*Omega*L/c)
    ARM = tArm * exp_2jOmegaL_c * np.array([[np.ones(nf), np.zeros(nf)], [-K, np.ones(nf)]])

    # the following code is generated by compile_ARM_RES_TF()
    # and is equivalent to the following:
    # RES = np.zeros((2,2,nf), dtype=complex)
    # RES = np.linalg.pinv(ID.transpose((2,0,1)) - rITM * ARM.transpose((2,0,1))).transpose((1,2,0))
    x0 = exp_2jOmegaL_c*rITM*tArm
    x1 = -x0 + 1
    x2 = 1/x1
    RES = np.array([[x2, np.zeros(nf)], [-K*x0/x1**2, x2]])
    # end of generated code

    rho_ARM = getProdTF(RES, (R + T) * ARM - rITM * ID)
    tau_ARM = tITM * RES

    # signal-extraction cavity
    SE = SEr * exp(1j * Omega * l / c)
    SEC = getProdTF(tSR * SE, getProdTF(rho_ARM, SE))

    exp_1jOmegal_c = exp(1j*Omega*l/c)
    r00 = RES[0,0,:]
    r10 = RES[1,0,:]
    r11 = RES[1,1,:]

    # the following code is generated by compile_SEC_RES_TF()
    # and is equivalent to the following:
    # RES = np.zeros((2,2,nf), dtype=complex)
    # RES = np.linalg.pinv(ID.transpose((2,0,1)) + rho * SEC.transpose((2,0,1))).transpose((1,2,0))
    x0 = cos(phi)
    x1 = exp_2jOmegaL_c*tArm*(R + T)
    x2 = -rITM + x1
    x3 = exp_1jOmegal_c**2*r11*tSR*x2
    x4 = sin(phi)
    x5 = exp_1jOmegal_c*r00*tSR*x2
    x6 = exp_1jOmegal_c*tSR*(-K*r11*x1 + r10*x2)
    x7 = exp_1jOmegal_c*(x0*x6 - x4*x5)
    x8 = rho*(x0**2*x3 + x4*x7) + 1
    x9 = x0*x3*x4
    x10 = exp_1jOmegal_c*(x0*x5 + x4*x6)
    x11 = x10*x4 + x9
    x12 = x0*x7 - x9
    x13 = rho*(x0*x10 - x3*x4**2) + 1
    x14 = 1/(-rho**2*x11*x12 + x13*x8)
    x15 = rho*x14
    RES = np.array([[x14*x8, -x11*x15], [-x12*x15, x13*x14]])
    # end of generated code

    rho_SEC = getProdTF(RES, RT_SRM * SEC + rho * ID)
    tau_SEC = tau * getProdTF(RES, SE)
    tau_SEC_ARM = getProdTF(tau_SEC, tau_ARM)

    # signal field
    Msig = tSig * exp(1j * Omega * L / c) * getProdTF(tau_SEC_ARM, np.array([[np.zeros(nf)], [sqrt(2 * K) / h_SQL]]))

    # dark-port input field
    Mifo = rho_SEC

    # loss field from arm cavity
    Mn = lossArm * tau_SEC_ARM

    # loss field from signal-extraction cavity
    Mp = lossSR * tau_SEC

    # adapt to GWINC phase convention
    Msig = Msig[[1, 0], :, :]
    Msig[1, 0, :] = -Msig[1, 0, :]

    def adapt_to_gwinc(Mx):
        My = np.zeros(Mx.shape, dtype=complex)
        My[0, 0, :] = Mx[1, 1, :]
        My[1, 1, :] = Mx[0, 0, :]
        My[0, 1, :] = -Mx[1, 0, :]
        My[1, 0, :] = -Mx[0, 1, :]
        return My

    Mifo = adapt_to_gwinc(Mifo)
    Mn = adapt_to_gwinc(Mn)
    Mp = adapt_to_gwinc(Mp)

    # overall coefficient
    coeff = 1

    # put all loss fields together
    Mnoise = np.hstack([Mn, Mp])

    return coeff, Mifo, Msig, Mnoise


def shotradSignalRecycledBnC(f, ifo, power):
    """Quantum noise model for signal recycled IFO

    See shotrad for more info.

    All references to Buonanno & Chen PRD 64 042006 (2001) (hereafter BnC)
    Updated to include losses DEC 2006 Kirk McKenzie using BnC notation
    Updated to include squeezing April 2009 KM
    Updated April 2010 KM, LB

    moved out of shotrad May 2010, mevans
    output is used in shotrad to compute final noise as
      n = coeff * (vHD * Msqz * Msqz' * vHD') / (vHD * Md * Md' * vHD')
    where
      Msqz = [Mc MsqueezeInput, Mn]

    coeff = frequency dependent overall noise coefficient (Nx1)
    Mifo = IFO input-output relation for the AS port
    Msig = signal transfer to the AS port
    Mnoise = noise fields produced by losses in the IFO at the AS port

    """
    lambda_ = ifo.Laser.Wavelength               # Laser Wavelength [m]
    hbar = const.hbar                            # Plancks Constant [Js]
    c = const.c                                  # SOL [m/s]
    Omega = 2*pi*f                               # [BnC, table 1] Signal angular frequency [rads/s]
    omega_0 = 2*pi*c/lambda_                     # [BnC, table 1] Laser angular frequency [rads/s]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    L = ifo.Infrastructure.Length                # Length of arm cavities [m]
    l = ifo.Optics.SRM.CavityLength              # SRC Length [m]
    T = ifo.Optics.ITM.Transmittance             # ITM Transmittance [Power]
    m = ifo.Suspension.Stage[0].Mass             # Mirror mass [kg]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    bsloss = ifo.Optics.BSLoss                   # BS Loss [Power]
    mismatch = 1 - ifo.Optics.coupling           # Mismatch
    mismatch = mismatch + ifo.TCS.SRCloss        # Mismatch

    # BSloss + mismatch has been incorporated into a SRC Loss
    lambda_SR = mismatch + bsloss                # SR cavity loss [Power]

    tau = sqrt(ifo.Optics.SRM.Transmittance)     # SRM Transmittance [amplitude]
    rho = sqrt(1 - tau**2 - lambda_SR)           # SRM Reflectivity [amplitude]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    ds = ifo.Optics.SRM.Tunephase                # SRC Detunning
    phi = (pi-ds)/2                              # [BnC, between 2.14 & 2.15] SR Detuning

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    lambda_arm = ifo.Optics.Loss*2               # [BnC, after 5.2] Round Trip loss in arm [Power]
    gamma_ac = T*c/(4*L)                         # [KLMTV-PRD2001] Arm cavity half bandwidth [1/s]
    epsilon = lambda_arm/(2*gamma_ac*L/c)        # [BnC, after 5.2] Loss coefficent for arm cavity

    I_0 = power.pbs                              # [BnC, Table 1] Power at BS (Power*prfactor) [W]
    I_SQL = (m*L**2*gamma_ac**4)/(4*omega_0)     # [BnC, 2.14] Power to reach free mass SQL
    Kappa = 2*((I_0/I_SQL)*gamma_ac**4)/ \
              (Omega**2*(gamma_ac**2+Omega**2))  # [BnC 2.13] Effective Radiation Pressure Coupling
    beta = arctan(Omega/gamma_ac)                # [BnC, after 2.11] Phase shift of GW SB in arm
    h_SQL = sqrt(8*hbar/(m*(Omega*L)**2))        # [BnC, 2.12] SQL Strain


    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Coefficients [BnC, Equations 5.8 to 5.12]
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    exp_1jbeta = exp(1j*beta)
    cos_beta = exp_1jbeta.real
    invexp_1jbeta = 1/exp_1jbeta
    exp_2jbeta = exp_1jbeta**2
    cos_2beta = exp_2jbeta.real
    invexp_2jbeta = 1/exp_2jbeta
    exp_4jbeta = exp_2jbeta**2
    C11_L = ( (1+rho**2) * ( cos(2*phi) + Kappa/2 * sin(2*phi) ) -
              2*rho*cos_2beta - 1/4*epsilon * ( -2 * (1+exp_2jbeta)**2 * rho + 4 * (1+rho**2) *
                                                cos_beta**2*cos(2*phi) + ( 3+exp_2jbeta ) *
                                                Kappa * (1+rho**2) * sin(2*phi) ) +
              lambda_SR * ( exp_2jbeta*rho-1/2 * (1+rho**2) * ( cos(2*phi)+Kappa/2 * sin(2*phi) ) ) )

    C22_L = C11_L

    C12_L = tau**2 * ( - ( sin(2*phi) + Kappa*sin(phi)**2 )+
                       1/2*epsilon*sin(phi) * ( (3+exp_2jbeta) * Kappa * sin(phi) + 4*cos_beta**2 * cos(phi)) +
                       1/2*lambda_SR * ( sin(2*phi)+Kappa*sin(phi)**2) )

    C21_L = tau**2 * ( (sin(2*phi)-Kappa*cos(phi)**2 ) +
                       1/2*epsilon*cos(phi) * ( (3+exp_2jbeta )*Kappa*cos(phi) - 4*cos_beta**2*sin(phi) ) +
                       1/2*lambda_SR * ( -sin(2*phi) + Kappa*cos(phi)**2) )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    D1_L = ( - (1+rho*exp_2jbeta ) * sin(phi) +
             1/4*epsilon * ( 3+rho+2*rho*exp_4jbeta + exp_2jbeta*(1+5*rho) ) * sin(phi)+
             1/2*lambda_SR * exp_2jbeta * rho * sin(phi) )

    D2_L = ( - (-1+rho*exp_2jbeta ) * cos(phi) +
             1/4*epsilon * ( -3+rho+2*rho*exp_4jbeta + exp_2jbeta * (-1+5*rho) ) * cos(phi)+
             1/2*lambda_SR * exp_2jbeta * rho * cos(phi) )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    P11 = 0.5 * sqrt(lambda_SR) * tau * \
          ( -2*rho*exp_2jbeta+2*cos(2*phi)+Kappa*sin(2*phi) )
    P22 = P11
    P12 = -sqrt(lambda_SR)*tau*sin(phi)*(2*cos(phi)+Kappa*sin(phi) )
    P21 =  sqrt(lambda_SR)*tau*cos(phi)*(2*sin(phi)-Kappa*cos(phi) )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # this was the PD noise source, but that belongs outside of this function
    #   I have used the equation for Q11 to properly normalize the other noises
    #   as well as the input-output relation Mc and the signal matrix Md

    Q11 = 1 / \
        ( invexp_2jbeta+rho**2*exp_2jbeta-rho*(2*cos(2*phi)+Kappa*sin(2*phi)) +
          1/2*epsilon*rho * (invexp_2jbeta*cos(2*phi)+exp_2jbeta*
                             ( -2*rho-2*rho*cos_2beta+cos(2*phi)+Kappa*sin(2*phi) ) +
                             2*cos(2*phi)+3*Kappa*sin(2*phi))-1/2*lambda_SR*rho *
          ( 2*rho*exp_2jbeta-2*cos(2*phi)-Kappa*sin(2*phi) ) )
    Q22 = Q11
    Q12 = 0
    Q21 = 0

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    N11 = sqrt(epsilon/2)*tau *(Kappa*(1+rho*exp_2jbeta)*sin(phi)+
                                2*cos_beta*(invexp_1jbeta*cos(phi)-rho*exp_1jbeta*(cos(phi)+Kappa*sin(phi))))
    N22 = -sqrt(2*epsilon)*tau*(-invexp_1jbeta+rho*exp_1jbeta)*cos_beta*cos(phi)
    N12 = -sqrt(2*epsilon)*tau*(invexp_1jbeta+rho*exp_1jbeta)*cos_beta*sin(phi);
    N21 = sqrt(epsilon/2)*tau*(-Kappa*(1+rho)*cos(phi)+
                               2*cos_beta*(invexp_1jbeta+rho*exp_1jbeta)*cos_beta*sin(phi))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # overall coefficient
    coeff = h_SQL**2/(2*Kappa*tau**2)

    # make normalization matrix
    Mq = make2x2TF(Q11, Q12, Q21, Q22)

    # 3D transfer matrices from vectors for each element
    Mifo = getProdTF(Mq, make2x2TF(C11_L, C12_L, C21_L, C22_L))
    Msig = getProdTF(Mq, np.array([D1_L, D2_L]).reshape(2, 1, f.size))

    # put all output noises together
    Mp = make2x2TF(P11, P12, P21, P22)
    Mn = make2x2TF(N11, N12, N21, N22)
    Mnoise = getProdTF(Mq, np.hstack((Mn, Mp)))

    return coeff, Mifo, Msig, Mnoise


def propagate_noise_fc_ifo(f, ifo, Mifo, Mnoise, sqz_params, alpha):
    """Propagate quantum noise through the filter cavities and IFO

    f: frequency vector [Hz]
    ifo: ifo Struct
    Mifo: IFO input-output relation for the AS port
    Mnoise: dictionary of existing noise sources
    sqz_params: Struct of squeeze parameters
    alpha: squeeze quadrature

    Returns a new Mnoise dict with the noises due to AS port vacuum, injection
    loss, and any input or output filter cavities added.
    """
    Mnoise = deepcopy(Mnoise)

    #>>>>>>>>    QUANTUM NOISE POWER SPECTRAL DENSITY WITH SQZ [BnC PRD 2004, 62]
    #<<<<<<<<<<<<<<<<< Modified to include losses (KM)
    #<<<<<<<<<<<<<<<<< Modified to include frequency dependent squeezing angle (LB)
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # ------------------------------------------- equation 63 BnC PRD 2004
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Define input matrix of Squeezing
    R = sqz_params.SQZ_DB / (20 * log10(exp(1)))           # squeeze factor
    R_anti = sqz_params.ANTISQZ_DB / (20 * log10(exp(1)))  # anti-squeeze factor
    Msqz = np.array([[exp(-R), 0], [0, exp(R_anti)]])

    # expand to Nfreq
    Msqz = np.transpose(np.tile(Msqz, (len(f), 1, 1)), axes=(1, 2, 0))

    # add input rotation
    MsqzRot = make2x2TF(cos(alpha), -sin(alpha), sin(alpha), cos(alpha))
    Msqz = getProdTF(MsqzRot, Msqz)
    Msqz = dict(ASvac=Msqz)

    # Include losses (lambda_in=ifo.Squeezer.InjectionLoss)
    Msqz = sqzInjectionLoss(Msqz, sqz_params.lambda_in, 'injection')

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Inject squeezed field into the IFO via some filter cavities
    if sqz_params.sqzType == 'Freq Dependent' and 'FilterCavity' in ifo.Squeezer:
        if not isinstance(ifo.Squeezer.FilterCavity, Sequence):
            fc_list = [ifo.Squeezer.FilterCavity]
        else:
            fc_list = ifo.Squeezer.FilterCavity
        logger.debug('  Applying %d input filter cavities' % len(fc_list))
        Mr, Msqz = sqzFilterCavityChain(
            f, fc_list, Msqz)

    # apply the IFO dependent squeezing matrix to get the total noise
    # due to quantum fluctuations coming in from the AS port
    Msqz = {key: getProdTF(Mifo, nn) for key, nn in Msqz.items()}

    # add this to the other noises
    Mnoise.update(Msqz)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # pass IFO output through some filter cavities
    if 'OutputFilter' in ifo:
        if ifo.OutputFilter.Type == 'None':
            # do nothing, say nothing
            pass

        elif ifo.OutputFilter.Type == 'Chain':
            logger.debug('  Applying %d output filter cavities' % np.atleast_1d(ifo.OutputFilter.FilterCavity).size)

            Mr, Mnoise = sqzFilterCavityChain(
                f, np.atleast_1d(ifo.OutputFilter.FilterCavity), Mnoise, key='OFC')
            Msig = getProdTF(Mr, Msig)
            #  Mnoise = getProdTF(Mn, Mnoise);

        elif ifo.OutputFilter.Type == 'Optimal':
            raise NotImplementedError("Cannot do optimal phase yet")
            logger.debug('  Optimal output filtering!')

            # compute optimal angle, including upcoming PD losses
            MnPD = sqzInjectionLoss(Mnoise, lambda_PD)
            zeta = sqzOptimalReadoutPhase(Msig, MnPD)

            # rotate by that angle, less the homodyne angle
            #zeta_opt = eta;
            cs = cos(zeta - eta)
            sn = sin(zeta - eta)
            Mrot = make2x2TF(cs, -sn, sn, cs)
            Mnoise = getProdTF(Mrot, Mnoise)
            Msig = getProdTF(Mrot, Msig)

        else:
            raise Exception('ifo.OutputFilter.Type must be None, Chain or Optimal, not "%s"' % ifo.OutputFilter.Type)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # add PD efficiency
    lambda_PD = 1 - ifo.Optics.PhotoDetectorEfficiency  # PD losses
    Mnoise = sqzInjectionLoss(Mnoise, lambda_PD, 'pd')

    return Mnoise


######################################################################
# Auxiliary functions
######################################################################

def make2x2TF(A11, A12, A21, A22):
    """Create transfer matrix with 2x2xnF.

    """
    A11, A12, A21, A22 = np.broadcast_arrays(A11, A12, A21, A22)
    M3 = np.array([[A11, A12], [A21, A22]])
    return M3.reshape(2, 2, -1)


def getProdTF(lhs, rhs):
    """Compute the product of M Nout x Nin x Naf frequency dependent transfer matrices

    See also getTF.

    NOTE: To perform more complicated operations on transfer
          matrices, see LTI object FRD ("help frd").  This
          function is the same as: freqresp(frd(lhs) * frd(rhs), f)

    """
    # check matrix size
    if lhs.shape[1] != rhs.shape[0]:
        raise Exception('Matrix size mismatch size(lhs, 2) = %d != %d = size(rhs, 1)' % (lhs.shape[1], rhs.shape[0]))
    if len(lhs.shape) == 3:
        lhs = np.transpose(lhs, axes=(2, 0, 1))
    if len(rhs.shape) == 3:
        rhs = np.transpose(rhs, axes=(2, 0, 1))

    # compute product
    if len(lhs.shape) < 3 or lhs.shape[0] == 1:
        rslt = np.matmul(lhs, rhs)
    elif len(rhs.shape) < 3 or rhs.shape[0] == 1:
        rslt = np.matmul(lhs, rhs)
    elif lhs.shape[0] == rhs.shape[0]:
        rslt = np.matmul(lhs, rhs)
    else:
        raise Exception('Matrix size mismatch lhs.shape[2] = %d != %d = rhs.shape[2]' % (lhs.shape[2], rhs.shape[2]))

    if len(rslt.shape) == 3:
        rslt = np.transpose(rslt, axes=(1, 2, 0))
    return rslt


def sqzInjectionLoss(Min, L, key):
    """Injection losses for squeezed field

    Calculates noise where unsqueezed vacuum is injected at a source of loss
    and adds this to a dictionary of existing noise sources

    Min = dictionary of existing sources
    L = the loss
    key = key for the new noise source
    Mout = updated dictionary with the original noises reduced by sqrt(1 - L)
           and a new item for the new noise source
    """
    # number of existing noise fields
    num_noise_fld = sum([nn.shape[1] for nn in Min.values()])
    # number of frequency points
    npts = Min[list(Min.keys())[0]].shape[2]
    # each of the existing noise sources should be reduced by sqrt(1 - L)
    Mout = {nkey: nn * sqrt(1 - L) for nkey, nn in Min.items()}

    # add new noise fields
    eye2 = np.eye(2, num_noise_fld)
    Meye = np.transpose(np.tile(eye2, (npts, 1, 1)), axes=(1, 2, 0))
    Mout[key] = Meye * sqrt(L)

    return Mout


def sqzFilterCavityChain(f, params, Mn, key='FC'):
    """Transfer relation for a chain of filter cavities

    Noise added by cavity losses are also output.

    f = frequency vector [Hz]
    param.fdetune = detuning [Hz]
    param.L = cavity length
    param.Ti = input mirror trasmission [Power]
    param.Li = input mirror loss
    param.Te = end mirror trasmission
    param.Le = end mirror loss
    param.Rot = phase rotation after cavity

    Mn0 = input noise
    Mc = input to output transfer
    Mn = filtered input noise, plus noise due to cavity losses
    key = key to use for new entries to the noise dictionary

    Note:
        [Mc, Mn] = sqzFilterCavityChain(f, params, Mn0)
      is the same as
        [Mc, Mn] = sqzFilterCavityChain(f, params);
        Mn = [getProdTF(Mc, Mn0), Mn];

    corresponding author: mevans

    """
    # make an identity TF
    Mc = make2x2TF(np.ones(f.shape), 0, 0, 1)

    # loop through the filter cavites
    for k, fc in enumerate(params):
        # extract parameters for this filter cavity
        Lf = fc.L
        fdetune = fc.fdetune
        Ti = fc.Ti
        Te = fc.Te
        Lrt = fc.Lrt
        theta = fc.Rot

        # compute new Mn
        fc_key = key + str(k)
        Mr, Mt, Mn = sqzFilterCavity(f, Lf, Ti, Te, Lrt, fdetune, Mn, key=fc_key)

        # apply phase rotation after filter cavity
        Mrot = np.array([[cos(theta), -sin(theta)], [sin(theta), cos(theta)]])
        Mn = {key: getProdTF(Mrot, nn) for key, nn in Mn.items()}

        # update Mc
        Mc = getProdTF(Mrot, getProdTF(Mr, Mc))

    return Mc, Mn


def sqzFilterCavity(f, Lcav, Ti, Te, Lrt, fdetune, MinR, MinT=1, key='FC'):
    """Reflection/transmission matrix for filter cavity

    Function which gives the reflection matrix for vacuum fluctuations
    entering the input mirror and the transmission matrix for vacuum
    fluctuations entering the end mirror of one filter cavity.  The
    input parameters are the cavity parameters and the 2X2 matrix of
    the incoming fields in the two-photon formalism.

    (R_alpha x S_r) for a freq independent squeezed field.
    f = vector frequency in Hz
    Lf = length of the filter cavity
    Ti = transmission and losses of the input mirror
    Te = transmission and losses of the end mirror
    Lrt = round-trip losses in the cavity (mirror transmissoins not included)
    fdetune: detuning frequency of the filter cavity [Hz]
    MinR: squeezed field injected from the input mirror of the filter cavity (a1,a2 basis)
         if this argument is empty, it is assumed that the user will use Mr,
         so no noise field is added to Mnoise.  If no argument is given, or
         the scalar 1 is given, an Mr unsqueezed input is assumed and Mr is
         concatenated into Mnoise.
    MinT: squeezed field injected from the back of the filter cavity
         with MinR, this argument can be omitted or set to 1 to indicate
         and unsqueezed input. [] can be used to avoid adding a noise
         term to Mnoise.
    key: key to use for new entries to the noise dictionary
         the losses from the filter cavity are key and the additional losses
         injected from the back mirror are key_unsqzd_back

    corresponding authors: LisaB, mevans

    """

    # reflectivities
    Ri = 1 - Ti
    Re = 1 - Te

    ri = sqrt(Ri)
    re = sqrt(Re)
    rr = ri * re * sqrt(1 - Lrt)  # include round-trip losses

    # Phases for positive and negative audio sidebands
    c = const.c
    omega = 2 * pi * f
    wf = 2 * pi * fdetune
    Phi_p = 2 * (omega-wf) * Lcav / c
    Phi_m = 2 * (-omega-wf) * Lcav / c

    ephi_p = exp(1j * Phi_p)
    ephi_m = exp(1j * Phi_m)

    # cavity gains
    g_p = 1 / (1 - rr * ephi_p)
    g_m = 1 / (1 - rr * ephi_m)

    # Reflectivity for vacuum flactuation entering the cavity from
    # the input mirror (check sign)
    r_p = ri - re * Ti * ephi_p * g_p
    r_m = ri - re * Ti * ephi_m * g_m

    # Transmissivity for vacuum flactuation entering the cavity from
    # the back mirror (check sign)
    t_p = sqrt(Ti * Te * ephi_p) * g_p
    t_m = sqrt(Ti * Te * ephi_m) * g_m

    # Transmissivity for vacuum flactuation entering the cavity from
    # the losses in the cavity
    l_p = sqrt(Ti * Lrt * ephi_p) * g_p
    l_m = sqrt(Ti * Lrt * ephi_m) * g_m

    # Relfection matrix for vacuum fluctuations entering from the input
    # mirror in the A+, (a-)* basis
    Mr_temp = make2x2TF(r_p, 0, 0, conj(r_m))

    # Transmission matrix for vacuum fluctuations entering from the end mirror
    Mt_temp = make2x2TF(t_p, 0, 0, conj(t_m))

    # Transmission matrix for vacuum fluctuations entering from the cavity losses
    Ml_temp = make2x2TF(l_p, 0, 0, conj(l_m))

    # Apply matrix which changes from two-photon basis to a+ and (a-)*
    Mbasis = np.array([[1, 1j], [1, -1j]])

    Mr = getProdTF(np.linalg.inv(Mbasis), getProdTF(Mr_temp, Mbasis))
    Mt = getProdTF(np.linalg.inv(Mbasis), getProdTF(Mt_temp, Mbasis))
    Ml = getProdTF(np.linalg.inv(Mbasis), getProdTF(Ml_temp, Mbasis))

    ###### output

    # reflected fields
    Mnoise = {}
    for nkey, nn in MinR.items():
        if np.isscalar(nn):
            Mnoise[nkey] = Mr * nn
        else:
            Mnoise[nkey] = getProdTF(Mr, nn)

    # transmitted fields
    if MinT != {} and Te > 0:
        if np.isscalar(MinT) and MinT == 1:
            # inject unsqueezed vacuum at the back
            Mnoise[key + '_unsqzd_back'] = Mt
        else:
            # inject a given noise at the back
            for nkey, nn in MinT.items():
                if np.isscalar(nn):
                    Mnoise[nkey] = Mt * nn
                else:
                    Mnoise[nkey] = getProdTF(Mt, nn)
    elif MinT != {} and Te == 0:
        Mnoise[key + '_unsqzd_back'] = np.zeros_like(Mt)

    # loss fields
    if Lrt > 0:
        Mnoise[key] = Ml

    return Mr, Mt, Mnoise



# import sys
# import versioneer
from setuptools import setup


# if {'build_sphinx'}.intersection(sys.argv):
#     setup_requires.append('sphinx')
# if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
#     setup_requires.append('pytest-runner')


setup(
    setup_requires=[
        'setuptools >= 30.3.0',
        'setuptools_scm',
    ],
    use_scm_version={
        'write_to': 'gwinc/_version.py',
    }
    # version=versioneer.get_version(),
    # cmdclass=versioneer.get_cmdclass(),
)



import argparse
import numpy as np

from . import IFOS, PLOT_STYLE
from .. import load_budget


FLO = 3
FHI = 10000
NPOINTS = 3000
YLIM = (1e-25, 1e-20)


def main():
    parser = argparse.ArgumentParser(
        description="Reference IFO comparison plot",
    )
    parser.add_argument(
        '--save', '-s',
        help="save plot to file (.pdf/.png/.svg)")
    args = parser.parse_args()

    if args.save:
        from matplotlib import use
        use('agg')
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)

    freq = np.logspace(np.log10(FLO), np.log10(FHI), NPOINTS)

    budgets = {}
    range_pad = 0
    for ifo in IFOS:
        budget = load_budget(ifo, freq)
        name = budget.name
        budgets[name] = budget
        range_pad = max(len(name), range_pad)

    for name, budget in budgets.items():
        trace = budget.run()
        try:
            import inspiral_range
            label_range = ' {:>6.0f} Mpc'.format(
                inspiral_range.range(freq, trace.psd),
            )
        except ModuleNotFoundError:
            label_range = ''

        label = '{name:<{pad}}{range}'.format(
            name=name,
            pad=range_pad,
            range=label_range,
        )
        ax.loglog(freq, trace.asd, label=label, linewidth=2)

    ax.grid(
        True,
        which='both',
        linewidth=0.5,
        ls='-',
        alpha=0.5,
    )

    ax.legend(
        fontsize='small',
        prop={'family': 'monospace'},
    )

    ax.autoscale(enable=True, axis='y', tight=True)
    ax.set_ylim(*YLIM)
    ax.set_xlim(freq[0], freq[-1])

    ax.set_xlabel('Frequency [Hz]')
    ax.set_ylabel(PLOT_STYLE['ylabel'])
    ax.set_title("PyGWINC reference IFO strain comparison")

    if args.save:
        fig.savefig(args.save)
    else:
        plt.show()


if __name__ == '__main__':
    main()



from gwinc.ifo.noises import *
from gwinc.ifo import PLOT_STYLE


class QuantumVacuum(nb.Budget):
    """Quantum Vacuum

    """
    style = dict(
        label='Quantum Vacuum',
        color='#ad03de',
    )

    noises = [
        QuantumVacuumAS,
        QuantumVacuumArm,
        QuantumVacuumSEC,
        QuantumVacuumFilterCavity,
        QuantumVacuumInjection,
        QuantumVacuumReadout,
        QuantumVacuumQuadraturePhase,
    ]


class Newtonian(nb.Budget):
    """Newtonian Gravity

    """

    name = 'Newtonian'

    style = dict(
        label='Newtonian',
        color='#15b01a',
    )

    noises = [
        NewtonianRayleigh,
        NewtonianBody,
        NewtonianInfrasound,
    ]


class Coating(nb.Budget):
    """Coating Thermal

    """

    name = 'Coating'

    style = dict(
        label='Coating Thermal',
        color='#fe0002',
    )

    noises = [
        CoatingBrownian,
        CoatingThermoOptic,
    ]


class Substrate(nb.Budget):
    """Substrate Thermal

    """

    name = 'Substrate'

    style = dict(
        label='Substrate Thermal',
        color='#fb7d07',
    )

    noises = [
        SubstrateBrownian,
        SubstrateThermoElastic,
    ]


class CE1(nb.Budget):

    name = 'Cosmic Explorer 1'

    noises = [
        QuantumVacuum,
        Seismic,
        Newtonian,
        SuspensionThermal,
        Coating,
        Substrate,
        ExcessGas,
    ]

    calibrations = [
        Strain,
    ]

    plot_style = PLOT_STYLE



from gwinc.ifo.noises import *
from gwinc.ifo import PLOT_STYLE


class QuantumVacuum(nb.Budget):
    """Quantum Vacuum

    """
    style = dict(
        label='Quantum Vacuum',
        color='#ad03de',
    )

    noises = [
        QuantumVacuumAS,
        QuantumVacuumArm,
        QuantumVacuumSEC,
        QuantumVacuumFilterCavity,
        QuantumVacuumInjection,
        QuantumVacuumReadout,
    ]


class Voyager(nb.Budget):

    name = 'Voyager'

    noises = [
        QuantumVacuum,
        Seismic,
        Newtonian,
        SuspensionThermal,
        CoatingBrownian,
        CoatingThermoOptic,
        ITMThermoRefractive,
        SubstrateBrownian,
        SubstrateThermoElastic,
        ExcessGas,
    ]

    calibrations = [
        Strain,
    ]

    plot_style = PLOT_STYLE



'''Functions to calculate coating thermal noise

'''
from __future__ import division, print_function
import numpy as np
from numpy import pi, exp, real, imag, sqrt, sin, cos, sinh, cosh, ceil, log

from .. import const
from ..const import BESSEL_ZEROS as zeta
from ..const import J0M as j0m


def coating_brownian(f, mirror, wavelength, wBeam, power=None):
    """Coating brownian noise for a given collection of coating layers

    This function calculates Coating Brownian noise using
    Hong et al . PRD 87, 082001 (2013).
    All references to 'the paper', 'Eq' and 'Sec' are to this paper.

    ***Important Note***
    Inside this function phi is used for denoting the phase shift suffered
    by light in one way propagation through a layer. This is in conflict
    with present nomenclature everywhere else where it is used as loss angle.

    The layers are assumed to be alernating low-n high-n layers, with
    low-n first.

    Inputs:
             f = frequency vector in Hz
        mirror = mirror properties Struct
    wavelength = laser wavelength
         wBeam = beam radius (at 1 / e**2 power)
         power = laser power falling on the mirror (W)

    If the power argument is present and is not None, the amplitude noise due
    to coating brownian noise will be calculated and its effect on the phase
    noise will be added (assuming the susceptibility is that of a free mass)

    ***The following parameters are experimental and unsupported as yet***
    The following optional parameters are available in the Materials object
    to provide separate Bulk and Shear loss angles and to include the effect
    of photoelasticity:
    lossBlown = Coating Bulk Loss Angle of Low Refrac.Index layer @ 100Hz
    lossSlown = Coating Shear Loss Angle of Low Refrac. Index layer @ 100Hz
    lossBhighn = Coating Bulk Loss Angle of High Refrac. Index layer @ 100Hz
    lossShighn = Coating Shear Loss Angle of High Refrac. Index layer @ 100Hz
    lossBlown_slope = Coating Bulk Loss Angle Slope of Low Refrac. Index layer
    lossSlown_slope = Coating Shear Loss Angle Slope of Low Refrac. Index layer
    lossBhighn_slope = Coating Bulk Loss Angle Slope of High Refrac. Index layer
    lossShighn_slope = Coating Shear Loss Angle Slope of High Refrac. Index layer
    PETlown = Relevant component of Photoelastic Tensor of High n layer*
    PEThighn = Relevant component of Photoelastic Tensor of Low n layer*

    Returns:
      SbrZ = Brownian noise spectra for one mirror in m**2 / Hz

    *
    Choice of PETlown and PEThighn can be inspired from sec. A.1. of the paper.
    There, values are chosen to get the longitudnal coefficent of
    photoelasticity as -0.5 for Tantala and -0.27 for Silica.
    These values also need to be added in Materials object.
    *
    If the optional arguments are not present, Phihighn and Philown will be
    used as both Bulk and Shear loss angles and PET coefficients will be set
    to 0.

    """
    # extract substructures
    sub = mirror.Substrate
    coat = mirror.Coating

    # Constants
    kBT = const.kB * sub.Temp
    c = const.c

    # substrate properties
    Ysub = sub.MirrorY            # Young's Modulous
    pratsub = sub.MirrorSigma     # Poisson Ratio
    nsub = sub.RefractiveIndex    # Refractive Index

    # coating properties
    dOpt = mirror.Coating.dOpt
    Yhighn = coat.Yhighn
    sigmahighn = coat.Sigmahighn
    nH = coat.Indexhighn
    if 'PEThighn' in coat:
        PEThighn = coat.PEThighn
    else:
        PEThighn = 0.0

    Ylown = coat.Ylown
    sigmalown = coat.Sigmalown
    nL = coat.Indexlown
    if 'PETlown' in coat:
        PETlown = coat.PETlown
    else:
        PETlown = 0.0

    # Loss Angles
    lossBhighn, lossShighn, lossBlown, lossSlown = interpretLossAngles(coat)

    # make vectors of material properties
    nol = len(dOpt)       # Number of layers
    nN = np.zeros(nol)
    yN = np.zeros(nol)
    pratN = np.zeros(nol)
    CPE = np.zeros(nol)

    # make simple alternating structure (low-index, high-index doublets)
    # (adapted from the more general calculation in
    #  Yam, W., Gras, S., & Evans, M. Multimaterial coatings with reduced
    #  thermal noise.
    #  Physical Review D, 91(4), 042002.  (2015).
    #  http://doi.org/10.1103/PhysRevD.91.042002 )
    nN[::2] = nL
    nN[1::2] = nH

    yN[::2] = Ylown
    yN[1::2] = Yhighn

    pratN[::2] = sigmalown
    pratN[1::2] = sigmahighn

    # Coefficient of photoelastic effect
    # PRD 87, 082001 Eq (A6)
    CPE[::2] = -0.5*PETlown*nL**3
    CPE[1::2] = -0.5*PEThighn*nH**3

    # geometrical thickness of each layer and total
    dGeo = wavelength * np.asarray(dOpt) / nN

    # WaveLength of light in each layer
    lambdaN = wavelength / nN

    # Calculate rho and derivatives of rho
    # with respect to both phi_k and r_j
    rho, dLogRho_dPhik, dLogRho_dRk, r = getCoatReflAndDer(nN, nsub, dOpt)

    # Define the function epsilon as per Eq (25)
    # Split epsilon function as:
    # Epsilon = Ep1 - Ep2 * cos(2k0n(z-zjp1)) - Ep3 * sin(2k0n(z-zjp1))

    # Part 1 of epsilon function
    Ep1 = (nN + CPE) * dLogRho_dPhik[:-1]
    # Part 2 of epsilon function (Prefactor of cosine term)
    Ep2 = CPE * (dLogRho_dPhik[:-1] * (1 - r[:-1]**2) / (2*r[:-1])
                 - (dLogRho_dPhik[1:] * (1 + r[:-1]**2) / (2 * r[:-1])))
    # Part 3 of epsilon function (Prefactor of sine term)
    Ep3 = (1 - r[:-1]**2) * CPE * dLogRho_dRk[:-1]

    # Define (1 - Im(epsilon)/2)
    Ip1 = 1 - imag(Ep1) / 2  # First part of (1 - Im(epsilon)/2)
    Ip2 = imag(Ep2) / 2      # Prefactor to cosine in (1 - Im(epsilon)/2)
    Ip3 = imag(Ep3) / 2      # Prefactor to sine in (1 - Im(epsilon)/2)

    # Define transfer functions from bulk and shear noise fields to layer
    # thickness and surface height as per Table I in paper
    C_B = np.sqrt(0.5*(1+pratN))
    C_SA = np.sqrt(1 - 2*pratN)
    D_B = ((1 - pratsub - 2*pratsub**2)*yN/(np.sqrt(2*(1+pratN))*Ysub))
    D_SA = -((1 - pratsub - 2*pratsub**2)*yN/(2*np.sqrt(1-2*pratN)*Ysub))
    D_SB = (np.sqrt(3)*(1-pratN)*(1 - pratsub - 2*pratsub**2)*yN
            / (2*np.sqrt(1-2*pratN)*(1+pratN)*Ysub))

    # Calculating effective beam area on each layer
    # Assuming the beam radius does not change significantly through the
    # depth of the mirror.
    Aeff = pi*(wBeam**2)

    # PSD at single layer with thickness equal to WaveLength
    # in the medium Eq (96)
    S_Bk = np.zeros((nol, len(f)))
    S_Bk[::2, :] = (4 * kBT * lambdaN[0] * lossBlown(f)
                    * (1 - pratN[0] - 2 * pratN[0]**2)
                    / (3 * pi * f * yN[0] * ((1 - pratN[0])**2) * Aeff))
    S_Bk[1::2, :] = (4 * kBT * lambdaN[1] * lossBhighn(f)
                     * (1 - pratN[1] - 2 * pratN[1]**2)
                     / (3 * pi * f * yN[1] * ((1 - pratN[1])**2) * Aeff))

    # Shear
    S_Sk = np.zeros((nol, len(f)))
    S_Sk[::2, :] = (4 * kBT * lambdaN[0] * lossSlown(f)
                    * (1 - pratN[0] - 2 * pratN[0]**2)
                    / (3 * pi * f * yN[0] * ((1 - pratN[0])**2) * Aeff))
    S_Sk[1::2, :] = (4 * kBT * lambdaN[1] * lossShighn(f)
                     * (1 - pratN[1] - 2 * pratN[1]**2)
                     / (3 * pi * f * yN[1] * ((1 - pratN[1])**2) * Aeff))

    # Coefficients q_j from Eq (94
    # See https://dcc.ligo.org/T2000552 for derivation
    k0 = 2 * pi / wavelength
    q_Bk = (+ 8 * C_B * (D_B + C_B * Ip1) * Ip3
            + 2 * C_B**2 * Ip2 * Ip3
            + 4 * (2 * D_B**2 + 4 * C_B * D_B * Ip1
                   + C_B**2 * (2 * Ip1**2 + Ip2**2 + Ip3**2)
                   ) * k0 * nN * dGeo
            - 8 * C_B * (D_B + C_B * Ip1) * Ip3 * np.cos(2 * k0 * nN * dGeo)
            - 2 * C_B**2 * Ip2 * Ip3 * np.cos(4 * k0 * nN * dGeo)
            + 8 * C_B * (D_B + C_B * Ip1) * Ip2 * np.sin(2 * k0 * nN * dGeo)
            + C_B**2 * (Ip2 - Ip3) * (Ip2 + Ip3) * np.sin(4 * k0 * nN * dGeo)
            ) / (8 * k0 * lambdaN * nN)

    q_Sk = (D_SB**2 * 8 * k0 * nN * dGeo
            + 8 * C_SA * (D_SA + C_SA * Ip1) * Ip3
            + 2 * C_SA**2 * Ip2 * Ip3
            + 4 * (2 * D_SA**2 + 4 * C_SA * D_SA * Ip1
                   + C_SA**2 * (2 * Ip1**2 + Ip2**2 + Ip3**2)
                   ) * k0 * nN * dGeo
            - 8 * C_SA * (D_SA + C_SA * Ip1) * Ip3 * np.cos(2 * k0 * nN * dGeo)
            - 2 * C_SA**2 * Ip2 * Ip3 * np.cos(4 * k0 * nN * dGeo)
            + 8 * C_SA * (D_SA + C_SA * Ip1) * Ip2 * np.sin(2 * k0 * nN * dGeo)
            + C_SA**2 * (Ip2 - Ip3) * (Ip2 + Ip3) * np.sin(4 * k0 * nN * dGeo)
            ) / (8 * k0 * lambdaN * nN)

    # S_Xi as per Eq(94)
    S_Xi = (np.tensordot(q_Bk, S_Bk, axes=1)
            + np.tensordot(q_Sk, S_Sk, axes=1))

    # From Sec II.E. Eq.(41)
    # Conversion of brownian amplitude noise to displacement noise
    if power is not None:

        # get/calculate optic transmittance
        mTi = mirror.get('Transmittance', 1-np.abs(rho)**2)

        # Define Re(epsilon)/2
        Rp1 = np.real(Ep1) / 2   # First part of Re(epsilon)/2
        Rp2 = -np.real(Ep2) / 2  # Prefactor to cosine in Re(epsilon)/2
        Rp3 = -np.real(Ep3) / 2  # Prefactor to sine in Re(epsilon)/2
        # Coefficients p_j from Eq (95)
        # See https://dcc.ligo.org/T2000552 for derivation
        p_BkbyC = (+ 8 * Rp1 * Rp3
                   + 2 * Rp2 * Rp3
                   + 4 * (2 * Rp1**2 + Rp2**2 + Rp3**2) * k0 * nN * dGeo
                   - 8 * Rp1 * Rp3 * np.cos(2 * k0 * nN * dGeo)
                   - 2 * Rp2 * Rp3 * np.cos(4 * k0 * nN * dGeo)
                   + 8 * Rp1 * Rp2 * np.sin(2 * k0 * nN * dGeo)
                   + (Rp2 - Rp3) * (Rp2 + Rp3) * np.sin(4 * k0 * nN * dGeo)
                   ) / (8 * k0 * lambdaN * nN)
        p_Bk = p_BkbyC * C_B**2
        p_Sk = p_BkbyC * C_SA**2

        # S_Zeta as per Eq(95)
        S_Zeta = (np.tensordot(p_Bk, S_Bk, axes=1)
                  + np.tensordot(p_Sk, S_Sk, axes=1))

        AmpToDispConvFac = ((32 * power)
                            / (mirror.MirrorMass * wavelength
                               * f**2 * c * 2 * pi * sqrt(mTi)))
        # Adding the two pathways of noise contribution as correlated noise
        SbrZ = (sqrt(S_Xi) + AmpToDispConvFac * sqrt(S_Zeta))**2
    else:
        SbrZ = S_Xi

    return SbrZ


def coating_thermooptic(f, mirror, wavelength, wBeam):
    """Optical coating thermo-optic displacement noise spectrum

    :f: frequency array in Hz
    :mirror: mirror parameter Struct
    :wavelength: laser wavelength
    :wBeam: beam radius (at 1 / e**2 power)

    :returns: tuple of:
    StoZ = displacement noise power spectrum at :f:
    SteZ = thermo-optic component of StoZ
    StrZ = thermo-refractive component of StoZ
    T = coating power transmission

    """
    # compute coefficients
    dTO, dTR, dTE, T, junk = getCoatTOPos(mirror, wavelength, wBeam)

    # compute correction factors
    gTO = getCoatThickCorr(f, mirror, wavelength, dTE, dTR)
    gTE = getCoatThickCorr(f, mirror, wavelength, dTE, 0)
    gTR = getCoatThickCorr(f, mirror, wavelength, 0, dTR)

    # compute thermal source spectrum
    SsurfT, junk = getCoatThermal(f, mirror, wBeam)

    StoZ = SsurfT * gTO * dTO**2
    SteZ = SsurfT * gTE * dTE**2
    StrZ = SsurfT * gTR * dTR**2

    return (StoZ, SteZ, StrZ, T)


def getCoatTOPos(mirror, wavelength, wBeam):
    """Mirror position derivative wrt thermal fluctuations

    :mirror: mirror parameter Struct
    :wavelength: laser wavelength
    :wBeam: beam radius (at 1 / e**2 power)

    :returns: tuple of:
    dTO = total thermo-optic dz/dT
    dTR = thermo-refractive dz/dT
    dTE = thermo-elastic dz/dT
    T = coating power transmission
    R = coating power reflection

    Compute thermal fluctuations with getCoatThermal.

    See LIGO-T080101.

    """
    # parameters
    nS = mirror.Substrate.RefractiveIndex
    dOpt = mirror.Coating.dOpt

    # compute refractive index, effective alpha and beta
    nLayer, aLayer, bLayer, dLayer, sLayer = getCoatLayers(mirror, wavelength)

    # compute coating average parameters
    dc, Cc, Kc, aSub = getCoatAvg(mirror, wavelength)

    # compute reflectivity and parameters
    dphi_dT, dphi_TE, dphi_TR, rCoat = getCoatTOPhase(1, nS, nLayer, dOpt, aLayer, bLayer, sLayer)
    R = abs(rCoat)**2
    T = 1 - R

    # for debugging
    #disp(sprintf('R = %.3f, T = %.0f ppm', R, 1e6 * T))

    # convert from phase to meters, subtracting substrate
    dTR = dphi_TR * wavelength / (4 * pi)
    dTE = dphi_TE * wavelength / (4 * pi) - aSub * dc

    # mirror finite size correction
    Cfsm = getCoatFiniteCorr(mirror, wavelength, wBeam)
    dTE = dTE * Cfsm

    # add TE and TR effects (sign is already included)
    dTO = dTE + dTR

    return dTO, dTR, dTE, T, R


def getCoatThickCorr(f, mirror, wavelength, dTE, dTR):
    """Finite coating thickness correction

    :f: frequency array in Hz
    :mirror: gwinc optic mirror structure
    :wavelength: laser wavelength
    :wBeam: beam radius (at 1 / e**2 power)

    Uses correction factor from LIGO-T080101, "Thick Coating
    Correction" (Evans).

    See getCoatThermoOptic for example usage.

    """
    ##############################################
    # For comparison in the bTR = 0 limit, the
    # equation from Fejer (PRD70, 2004)
    # modified so that gFC -> 1 as xi -> 0
    #  gTC = (2 ./ (R * xi.**2)) .* (sh - s + R .* (ch - c)) ./ ...
    #    (ch + c + 2 * R * sh + R**2 * (ch - c));
    # which in the limit of xi << 1 becomes
    #  gTC = 1 - xi * (R - 1 / (3 * R));

    # parameter extraction
    pS = mirror.Substrate
    Cs = pS.MassCM * pS.MassDensity
    Ks = pS.MassKappa

    # compute coating average parameters
    dc, Cc, Kc, junk = getCoatAvg(mirror, wavelength)

    # R and xi (from T080101, Thick Coating Correction)
    w = 2 * pi * f
    R = sqrt(Cc * Kc / (Cs * Ks))
    xi = dc * sqrt(2 * w * Cc / Kc)

    # trig functions of xi
    s = sin(xi)
    c = cos(xi)
    sh = sinh(xi)
    ch = cosh(xi)

    # pR and pE (dTR = -\bar{\beta} lambda, dTE = \Delta \bar{\alpha} d)
    pR = dTR / (dTR + dTE)
    pE = dTE / (dTR + dTE)

    # various parts of gTC
    g0 = 2 * (sh - s) + 2 * R * (ch - c)
    g1 = 8 * sin(xi / 2) * (R * cosh(xi / 2) + sinh(xi / 2))
    g2 = (1 + R**2) * sh + (1 - R**2) * s + 2 * R * ch
    gD = (1 + R**2) * ch + (1 - R**2) * c + 2 * R * sh

    # and finally, the correction factor
    gTC = (pE**2 * g0 + pE * pR * xi * g1 + pR**2 * xi**2 * g2) / (R * xi**2 * gD)

    return gTC


def getCoatThermal(f, mirror, wBeam):
    """Thermal noise spectra for a surface layer

    :f: frequency array in Hz
    :mirror: mirror parameter Struct
    :wBeam: beam radius (at 1 / e**2 power)

    :returns: tuple of:
    SsurfT = power spectra of thermal fluctuations in K**2 / Hz
    rdel = thermal diffusion length at each frequency in m

    """
    pS = mirror.Substrate
    C_S = pS.MassCM * pS.MassDensity
    K_S = pS.MassKappa
    kBT2 = const.kB * pS.Temp**2

    # omega
    w = 2 * pi * f

    # thermal diffusion length
    rdel = sqrt(2 * K_S / (C_S * w))

    # noise equation
    SsurfT = 4 * kBT2 / (pi * w * C_S * rdel * wBeam**2)

    return SsurfT, rdel


def getCoatLayers(mirror, wavelength):
    """Layer vectors for refractive index, effective alpha and beta and geometrical thickness

    :mirror: mirror parameter Struct
    :wavelength: laser wavelength

    :returns: tuple of:
    nLayer = refractive index of each layer, ordered input to output (N x 1)
    aLayer = change in geometrical thickness with temperature
           = the effective thermal expansion coeffient of the coating layer
    bLayer = change in refractive index with temperature
           = dn/dT
    dLayer = geometrical thicness of each layer
    sLayer = Yamamoto thermo-refractive correction
           = alpha * (1 + sigma) / (1 - sigma)

    """
    # coating parameters
    pS = mirror.Substrate
    pC = mirror.Coating
    dOpt = mirror.Coating.dOpt

    Y_S = pS.MirrorY
    sigS = pS.MirrorSigma

    alphaL = pC.Alphalown
    betaL = pC.Betalown
    Y_L = pC.Ylown
    sigL = pC.Sigmalown
    nL = pC.Indexlown

    alphaH = pC.Alphahighn
    betaH = pC.Betahighn
    Y_H = pC.Yhighn
    sigH = pC.Sigmahighn
    nH = pC.Indexhighn

    Nlayer = len(dOpt)

    # compute effective alpha
    def getExpansionRatio(Y_C, sigC, Y_S, sigS):
        ##############################################
        # Y_C and sigC are for the coating material (can also be substrate)
        # Y_S and sigS are for the substrate material
        #
        ce = ((1 + sigS) / (1 - sigC)) \
             * ( ((1 + sigC) / (1 + sigS)) + (1 - 2 * sigS) * Y_C / Y_S )
        return ce

    aLayer = np.zeros(Nlayer)
    aLayer[::2] = alphaL * getExpansionRatio(Y_L, sigL, Y_S, sigS)
    aLayer[1::2] = alphaH * getExpansionRatio(Y_H, sigH, Y_S, sigS)

    # and beta
    bLayer = np.zeros(Nlayer)
    bLayer[::2] = betaL
    bLayer[1::2] = betaH

    # and refractive index
    nLayer = np.zeros(Nlayer)
    nLayer[::2] = nL
    nLayer[1::2] = nH

    # and geometrical thickness
    dLayer = wavelength * np.asarray(dOpt) / nLayer

    # and sigma correction
    sLayer = np.zeros(Nlayer)
    sLayer[::2] = alphaL * (1 + sigL) / (1 - sigL)
    sLayer[1::2] = alphaH * (1 + sigH) / (1 - sigH)

    return nLayer, aLayer, bLayer, dLayer, sLayer


def getCoatAvg(mirror, wavelength):
    """Coating average properties

    :mirror: gwinc optic mirror structure
    :wavelength: laser wavelength

    :returns: tuple of:
    dc = total thickness (meters)
    Cc = heat capacity
    Kc = thermal diffusivity
    aSub = effective substrate thermal expansion (weighted by heat capacity)

    """
    # coating parameters
    pS = mirror.Substrate
    pC = mirror.Coating
    dOpt = mirror.Coating.dOpt

    alphaS = pS.MassAlpha
    C_S = pS.MassCM * pS.MassDensity
    sigS = pS.MirrorSigma

    C_L = pC.CVlown
    K_L = pC.ThermalDiffusivitylown

    C_H = pC.CVhighn
    K_H = pC.ThermalDiffusivityhighn

    # compute refractive index, effective alpha and beta
    junk1, junk2, junk3, dLayer, junk4 = getCoatLayers(mirror, wavelength)

    # heat capacity
    dc = np.sum(dLayer)
    dL = np.sum(dLayer[::2])
    dH = np.sum(dLayer[1::2])
    Cc = (C_L * dL + C_H * dH) / dc

    # thermal diffusivity
    KinvL = 1 / K_L
    KinvH = 1 / K_H
    Kc = dc / (KinvL * dL + KinvH * dH)

    # effective substrate thermal expansion
    aSub = 2 * alphaS * (1 + sigS) * Cc / C_S

    return dc, Cc, Kc, aSub


def getCoatTOPhase(nIn, nOut, nLayer, dOpt, aLayer, bLayer, sLayer):
    """Coating reflection phase derivatives w.r.t. temperature

    :nIn: refractive index of input medium (e.g., vacuum = 1)
    :nOut: refractive index of output medium (e.g., SiO2 = 1.45231 @ 1064nm)
    :nLayer: refractive index of each layer, ordered input to output (N x 1)
    :dOpt: optical thickness / lambda of each layer
           = geometrical thickness * refractive index / lambda
    :aLayer: change in geometrical thickness with temperature
             = the effective thermal expansion coeffient of the coating layer
    :bLayer: change in refractive index with temperature
             = dn/dT
             = dd/dT - n * a

    :returns: tuple of:
    dphi_dT = total thermo-optic phase derivative with respect to temperature
            = dphi_TE + dphi_TR
    dphi_TE = thermo-elastic phase derivative (dphi / dT)
    dphi_TR = thermo-refractive phase derivative (dphi / dT)
    rCoat = amplitude reflectivity of coating (complex)

    Note about aLayer: on a SiO2 substrate,
    a_Ta2O5 ~ 3.5 * alpha_Ta2O5
    a_SiO2 ~ 2.3 * alpha_SiO2

    See :getCoatTOPos: for more information.

    See LIGO-T080101.

    """
    # coating reflectivity calc
    rCoat, dcdp = getCoatRefl2(nIn, nOut, nLayer, dOpt)[:2]

    # geometrical distances
    dGeo = np.asarray(dOpt) / nLayer

    # phase derivatives
    dphi_dd = 4 * pi * dcdp

    # thermo-refractive coupling
    dphi_TR = np.sum(dphi_dd * (bLayer + sLayer * nLayer) * dGeo)

    # thermo-elastic
    dphi_TE = 4 * pi * np.sum(aLayer * dGeo)

    # total
    dphi_dT = dphi_TR + dphi_TE

    return dphi_dT, dphi_TE, dphi_TR, rCoat


def getCoatFiniteCorr(mirror, wavelength, wBeam):
    """Finite mirror size correction

    :mirror: mirror parameter Struct
    :wavelength: laser wavelength
    :wBeam: beam radius (at 1 / e**2 power)

    Uses correction factor from PLA 2003 vol 312 pg 244-255
    "Thermodynamical fluctuations in optical mirror coatings"
    by V. B. Braginsky and S. P. Vyatchanin
    http://arxiv.org/abs/cond-mat/0302617

    (see getCoatTOPos for example usage)

    version 1 by Sam Wald, 2008

    """
    # parameter extraction
    R = mirror.MassRadius
    H = mirror.MassThickness
    dOpt = mirror.Coating.dOpt

    alphaS = mirror.Substrate.MassAlpha
    C_S = mirror.Substrate.MassCM * mirror.Substrate.MassDensity
    Y_S = mirror.Substrate.MirrorY
    sigS = mirror.Substrate.MirrorSigma

    alphaL = mirror.Coating.Alphalown
    C_L = mirror.Coating.CVlown
    Y_L = mirror.Coating.Ylown
    sigL = mirror.Coating.Sigmalown
    nL = mirror.Coating.Indexlown

    alphaH = mirror.Coating.Alphahighn
    C_H = mirror.Coating.CVhighn
    Y_H = mirror.Coating.Yhighn
    sigH = mirror.Coating.Sigmahighn
    nH = mirror.Coating.Indexhighn

    # coating sums
    dL = wavelength * np.sum(dOpt[::2]) / nL
    dH = wavelength * np.sum(dOpt[1::2]) / nH
    dc = dH + dL

    # AVERAGE SPECIFIC HEAT (simple volume average for coating)
    Cf = (C_L * dL + C_H * dH) / dc
    Cr = Cf / C_S

    # COATING AVERAGE VALUE X = ALPHAF*(1+POISSONf)/(1-POISSONf) avg
    xxL = alphaL * (1 + sigL) / (1 - sigL)
    xxH = alphaH * (1 + sigH) / (1 - sigH)
    Xf = (xxL * dL + xxH * dH) / dc
    Xr = Xf / alphaS

    # COATING AVERAGE VALUE Y = ALPHAF* YOUNGSF/(1-POISSONF) avg
    yyL = alphaL * Y_L / (1 - sigL)
    yyH = alphaH * Y_H / (1 - sigH)
    Yf = (yyL * dL + yyH * dH) / dc
    Yr = Yf / (alphaS * Y_S)

    # COATING AVERAGE VALUE Z = 1/(1-POISSONF) avg
    zzL = 1 / (1 - sigL)
    zzH = 1 / (1 - sigH)
    Zf = (zzL * dL + zzH * dH) / dc

    #################################### FINITE SIZE CORRECTION CALCULATION

    # beam size parameter used by Braginsky
    r0 = wBeam / sqrt(2)

    # between eq 77 and 78
    km = zeta / R
    Qm = exp(-2 * km * H)
    pm = exp(-km**2 * r0**2 / 4) / j0m # left out factor of pi * R**2 in denominator

    # eq 88
    Lm = Xr - Zf * (1 + sigS) + (Yr * (1 - 2 * sigS) + Zf - 2 * Cr) * \
         (1 + sigS) * (1 - Qm)**2 / ((1 - Qm)**2 - 4 * km**2 * H**2 * Qm)

    # eq 90 and 91
    S1 = (12 * R**2 / H**2) * np.sum(pm / zeta**2)
    S2 = np.sum(pm**2 * Lm**2)
    P = (Xr - 2 * sigS * Yr - Cr + S1 * (Cr - Yr * (1 - sigS)))**2 + S2

    # eq 60 and 70
    LAMBDA = -Cr + (Xr / (1 + sigS) + Yr * (1 - 2 * sigS)) / 2

    # eq 92
    Cfsm = sqrt((r0**2 * P) / (2 * R**2 * (1 + sigS)**2 * LAMBDA**2))

    return Cfsm


def getCoatDopt(materials, T, dL, dCap=0.5):
    """Coating layer optical thicknesses to match desired transmission

    :materials: gwinc optic materials structure
    :T: power transmission of coating
    :dL: optical thickness of low-n layers (high-n layers have dH = 0.5 - dL)
    :dCap: first layer (low-n) thickness (default 0.5)

    :returns: optical thickness array Nlayer x 1 (dOpt)

    """
    ##############################################
    def getTrans(materials, Ndblt, dL, dH, dCap, dTweak):

        # the optical thickness vector
        dOpt = np.zeros(2 * Ndblt)
        dOpt[0] = dCap
        dOpt[1::2] = dH
        dOpt[2::2] = dL

        N = dTweak.size
        T = np.zeros(N)
        for n in range(N):
            dOpt[-1] = dTweak[n]
            r = getCoatRefl(materials, dOpt)[0]
            T[n] = 1 - abs(r**2)

        return T

    ##############################################
    def getTweak(materials, T, Ndblt, dL, dH, dCap, dScan, Nfit):

        # tweak bottom layer
        Tn = getTrans(materials, Ndblt, dL, dH, dCap, dScan)
        pf = np.polyfit(dScan, Tn - T, Nfit)
        rts = np.roots(pf)
        if not any((imag(rts) == 0) & (rts > 0)):
            dTweak = None
            Td = 0
            return dTweak, Td
        dTweak = real(np.min(rts[(imag(rts) == 0) & (rts > 0)]))

        # compute T for this dTweak
        Td = getTrans(materials, Ndblt, dL, dH, dCap, np.array([dTweak]))

        return dTweak, Td

        # plot for debugging
        #   plot(dScan, [Tn - T, polyval(pf, dScan)], dTweak, Td - T, 'ro')
        #   grid on
        #   legend('T exact', 'T fit', 'T at dTweak')
        #   title(sprintf('%d doublets', Ndblt))
        #   pause(1)

    # get IFO model stuff (or create it for other functions)
    pS = materials.Substrate
    pC = materials.Coating

    nS = pS.RefractiveIndex
    nL = pC.Indexlown
    nH = pC.Indexhighn

    ########################
    # find number of quarter-wave layers required, as first guess
    nR = nH / nL
    a1 = (2 - T + 2 * sqrt(1 - T)) / (nR * nH * T)
    Ndblt = int(ceil(log(a1) / (2 * log(nR))))

    # search through number of doublets to find Ndblt
    # which gives T lower than required
    dH = 0.5 - dL
    Tn = getTrans(materials, Ndblt, dL, dH, dCap, np.array([dH]))
    while Tn < T and Ndblt > 1:
        # strange, but T is too low... remove doublets
        Ndblt = Ndblt - 1
        Tn = getTrans(materials, Ndblt, dL, dH, dCap, np.array([dH]))
    while Tn > T and Ndblt < 1e3:
        # add doublets until T > tN
        Ndblt = Ndblt + 1
        Tn = getTrans(materials, Ndblt, dL, dH, dCap, np.array([dH]))

    ########################
    # tweak bottom layer
    delta = 0.01
    dScan = np.arange(0, 0.25+delta, delta)
    dTweak = getTweak(materials, T, Ndblt, dL, dH, dCap, dScan, 5)[0]

    if not dTweak:
        if nS > nL:
            raise Exception('Coating tweak layer not sufficient since nS > nL.')
        else:
            raise Exception('Coating tweak layer not found... very strange.')

    # now that we are close, get a better result with a linear fit
    delta = 0.001
    dScan = np.linspace(dTweak - 3*delta, dTweak + 3*delta, 7)
    dTweak, Td = getTweak(materials, T, Ndblt, dL, dH, dCap, dScan, 3)

    # negative values are bad
    if dTweak < 0.01:
        dTweak = 0.01

    # check the result
    if abs(log(Td / T)) > 1e-3:
        print('Exact coating tweak layer not found... %g%% error.' % abs(log(Td / T)))

    ########################
    # return dOpt vector
    dOpt = np.zeros(2 * Ndblt)
    dOpt[0] = dCap
    dOpt[1::2] = dH
    dOpt[2::2] = dL
    dOpt[-1] = dTweak

    return dOpt


def getCoatRefl(materials, dOpt):
    """Amplitude reflectivity, with phase, of a coating

    :materials: gwinc optic materials sturcutre
    :dOpt: coating layer thickness array (Nlayer x 1)

    :returns: see return value of :geteCoatRefl2:

    """
    pS = materials.Substrate
    pC = materials.Coating

    nS = pS.RefractiveIndex
    nL = pC.Indexlown
    nH = pC.Indexhighn

    Nlayer = len(dOpt)

    # refractive index of input, coating, and output materials
    nAll = np.zeros(Nlayer + 2)
    nAll[0] = 1  # vacuum input
    nAll[1::2] = nL
    nAll[2::2] = nH
    nAll[-1] = nS # substrate output

    # backend calculation
    return getCoatRefl2(nAll[0], nAll[-1], nAll[1:-1], dOpt)


def getCoatRefl2(nIn, nOut, nLayer, dOpt):
    """Coating reflection and phase derivatives

    :nIn: refractive index of input medium (e.g., vacuum = 1)
    :nOut: refractive index of output medium (e.g., SiO2 = 1.45231 @ 1064nm)
    :nLayer: refractive index of each layer, ordered input to output (N x 1)
    :dOpt: optical thickness / lambda of each layer,
           geometrical thickness * refractive index / lambda

    :returns: tuple of:
    rCoat = amplitude reflectivity of coating (complex) = rbar(0)
    dcdp = d reflection phase / d round-trip layer phase
    rbar = amplitude reflectivity of coating from this layer down
    r = amplitude reflectivity of this interface (r(1) is nIn to nLayer(1))

    See LIGO-T080101.

    """
    # Z-dir (1 => away from the substrate, -1 => into the substrate)
    zdir = 1

    # vector of all refractive indexs
    nAll = np.concatenate(([nIn], nLayer, [nOut]))

    # reflectivity of each interface
    r = (nAll[:-1] - nAll[1:]) / (nAll[:-1] + nAll[1:])

    # combine reflectivities
    rbar = np.zeros(r.size, dtype=complex)
    ephi = np.zeros(r.size, dtype=complex)

    # round-trip phase in each layer
    ephi[0] = 1
    ephi[1:] = exp(4j * zdir * pi * np.asarray(dOpt))

    rbar[-1] = ephi[-1] * r[-1]
    for n in range(len(dOpt), 0, -1):
        # accumulate reflectivity
        rbar[n-1] = ephi[n-1] * (r[n-1] + rbar[n]) / (1 + r[n-1] * rbar[n])

    # reflectivity derivatives
    dr_dphi = ephi[:-1] * (1 - r[:-1]**2) / (1 + r[:-1] * rbar[1:])**2
    dr_dphi = (1j * zdir * rbar[1:]) * np.multiply.accumulate(dr_dphi)

    # shift rbar index
    rCoat = rbar[0]
    rbar = rbar[1:]

    # phase derivatives
    dcdp = -imag(dr_dphi / rCoat)  ### Where did this minus come from???

    return rCoat, dcdp, rbar, r


def getCoatReflAndDer(nN, nsub, dOpt):
    '''
    Helper function for coating_brownian_hong().
    Follows Hong et al . PRD 87, 082001 (2013) Sec V.A.
    This function calculates derivatives of complex reflectivity of Coating
    with respect to phase shifts through each layer and reflectivities of
    each interface
    Input:

      nN = Refractive indices of coatings layers
    nsub = Refractive Index of Substrate
    dOpt = optical thickness / lambda of each layer,
           geometrical thickness * refractive index / lambda

    Returns:
     delLogRho_delPhik = Partial derivative of log of total effective
                         reflectivity of coating with respect to phase shifts
                         in each layer.
    delLogRho_delReflk = Partial derivative of log of total effective
                         reflectivity of coating with respect to reflectivity
                         of each interface.
    '''
    nol = len(dOpt)  # Number of layers in coating
    # Reflectivities and transmitivities
    # r[j] is reflectivity from (j-1)th and (j)th layer interface
    # Here r[0] is reflectivity from air and 0th layer
    # and r[-1] is reflectivity between last layer and substrate
    Refl = np.zeros(nol+1)
    Refl[0] = (1 - nN[0]) / (1 + nN[0])
    Refl[1:-1] = (nN[:-1] - nN[1:]) / (nN[:-1] + nN[1:])
    Refl[-1] = (nN[-1] - nsub) / (nN[-1] + nsub)
    # Note the shift from nomenclature
    # Phi is reserved for denoting one-way phase shift suffered by light
    # during propagation through a layer
    Phi = np.asarray(dOpt) * 2 * pi

    # Define rho_k as reflectivity of
    # k layers starting from (N-k-1)th lyer to (N-1)th layer
    # So rhoN[-1] is reflectivity for  no layers but interface from
    #                                              last layer to substrate
    # rhoN[0] is total complex reflectivity of the coating stack.
    rhoN = np.zeros_like(Refl, np.complex128)

    phiNmkm1 = np.flipud(Phi)             # phi_{N-k-1}
    rNmkm1 = np.flipud(Refl[:-1])              # r_{N-k-1}
    exp2iphiNmkm1 = np.exp(2j*phiNmkm1)      # exp(2i phi_{N-k-1})

    # Recursion relation for complex reflectivity
    # See https://dcc.ligo.org/T2000552 for derivation
    rhoN[0] = Refl[-1]
    for k in range(len(Refl)-1):
        rhoN[k+1] = ((rNmkm1[k] + exp2iphiNmkm1[k] * rhoN[k])
                     / (1 + exp2iphiNmkm1[k] * rNmkm1[k] * rhoN[k]))

    denTerm = (1 + exp2iphiNmkm1 * rNmkm1 * rhoN[:-1])**2

    # Derivatives of rho_{k+1} wrt to rho_{k}, r_{N-k-1} and phi_{N-k-1}
    delRhokp1_delRhok = exp2iphiNmkm1 * (1 - rNmkm1**2) / denTerm
    delRhokp1_delRNmkm1 = np.append(1, ((1 - (exp2iphiNmkm1*rhoN[:-1])**2)
                                        / denTerm))
    delRhokp1_delPhiNmkm1 = np.append(0, -2j * rhoN[:-1] * delRhokp1_delRhok)

    # Derivative of rho_{N} wrt to rho_{N-j}
    delRhoN_delRhoNmj = np.append(1, np.cumprod(np.flipud(delRhokp1_delRhok)))

    # Derivative of rho_{N} wrt to r_k and phi_k
    delRho_delRk = - delRhoN_delRhoNmj * np.flipud(delRhokp1_delRNmkm1)
    delRho_delPhik = - delRhoN_delRhoNmj * np.flipud(delRhokp1_delPhiNmkm1)
    delLogRho_delReflk = delRho_delRk / rhoN[-1]
    delLogRho_delPhik = delRho_delPhik / rhoN[-1]
    delLogRho_delPhik[-1] = 0        # Define this as per Eq (26)

    return rhoN[-1], delLogRho_delPhik, delLogRho_delReflk, Refl


def interpretLossAngles(coat):
    '''
    Helper function for coating_brownian().

    Creates function from 100 Hz value of loss angle and is logarithmic
    slope.

    if separate bulk and shear loss angles are not provided as
    lossBhighn, lossShighn, lossBlown and lossSlown
    then earlier version names of Phihighn and Philown are searched for and
    used as Bulk loss angle while setting Shear loss angles to zero.

    Input argument:
    coat = Coating object containing loss angle values or expressions.
    Returns:
    lossBlown = Coating Bulk Loss Angle of Low Refractive Index layer
    lossSlown = Coating Shear Loss Angle of Low Refractive Index layer
    lossBhighn = Coating Bulk Loss Angle of High Refractive Index layer
    lossShighn = Coating Shear Loss Angle of High Refractive Index layer
    '''
    if 'lossBhighn' in coat and 'lossShighn' in coat:
        if 'lossBhighn_slope' in coat:
            def lossBhighn(f):
                return coat.lossBhighn * (f / 100)**coat.lossBhighn_slope
        else:
            def lossBhighn(f): return coat.lossBhighn
        if 'lossShighn_slope' in coat:
            def lossShighn(f):
                return coat.lossShighn * (f / 100)**coat.lossShighn_slope
        else:
            def lossShighn(f): return coat.lossShighn
    else:
        # Use Phihighn if specific Bulk & Shear loss angles not provided
        if 'Phihighn_slope' in coat:
            def Phihighn(f):
                return coat.Phihighn * (f / 100)**coat.Phihighn_slope
            lossBhighn = lossShighn = Phihighn
        else:
            lossBhighn = lossShighn = lambda f: coat.Phihighn

    if 'lossBlown' in coat and 'lossSlown' in coat:
        if 'lossBlown_slope' in coat:
            def lossBlown(f):
                return coat.lossBlown * (f / 100)**coat.lossBlown_slope
        else:
            def lossBlown(f): return coat.lossBlown
        if 'lossSlown_slope' in coat:
            def lossSlown(f):
                return coat.lossSlown * (f / 100)**coat.lossSlown_slope
        else:
            def lossSlown(f): return coat.lossSlown
    else:
        # Use Philown if specific Bulk & Shear loss angles not provided
        if 'Philown_slope' in coat:
            def Philown(f):
                return coat.Philown * (f / 100)**coat.Philown_slope
            lossBlown = lossSlown = Philown
        else:
            lossBlown = lossSlown = lambda f: coat.Philown

    return lossBhighn, lossShighn, lossBlown, lossSlown



import os
import numpy as np

from .. import IFOS, load_budget
from .. import io


def test_io(tmpdir):
    freq = np.logspace(
        np.log10(5),
        np.log10(6000),
        1000,
    )
    budget = load_budget(IFOS[0], freq)
    traces = budget.run()
    path = os.path.join(tmpdir, 'foo.h5')
    io.save_hdf5(path, traces, ifo=budget.ifo)
    otraces = io.load_hdf5(path)
    assert np.all(traces.freq == otraces.freq)
    assert np.all(traces.psd == otraces.psd)
    for name, trace in otraces.items():
        assert np.all(traces[name].freq == trace.freq)
        assert np.all(traces[name].psd == trace.psd)
    assert budget.ifo == otraces.ifo



"""
"""

from collections.abc import Mapping, Sequence, MutableSequence
from numbers import Number

import os
import re
import io
import yaml

import numpy as np
from scipy.io import loadmat
from scipy.io.matlab.mio5_params import mat_struct


class Struct(object):
    """Matlab struct-like object

    This is a simple implementation of a MATLAB struct-like object
    that stores values as attributes of a simple class: and allows
    assigning to attributes recursively, e.g.:

    >>> s = Struct()
    >>> s.a = 4
    >>> s.b = Struct()
    >>> s.b.c = 8

    Various classmethods allow creating one of these objects from YAML
    file, a nested dict, or a MATLAB struct object.

    """
    STRUCT_EXT = ['.yaml', '.yml', '.mat', '.m']
    """accepted extension types for struct files"""

    # FIXME: There should be a way to allow setting nested struct
    # attributes, e.g.:
    #
    # >>> s = Struct()
    # >>> s.a.b.c = 4
    #
    # Usage of __getattr__ like this is dangerous and creates
    # non-intuitive behavior (i.e. an empty struct is returned when
    # accessing attributes that don't exist).  Is there a way to
    # accomplish this without that adverse side affect?
    #
    # def __getattr__(self, name):
    #     if name not in self.__dict__:
    #         self.__dict__[name] = Struct()
    #     return self.__dict__[name]

    ##########

    def __init__(self, *args, **kwargs):
        """Initialize Struct object

        Initializes similar to dict(), taking a single dict or mapping
        argument, or keyword arguments to initially populate the
        Struct.

        """
        # TODO, should this use the more or less permissive allow_unknown_types?
        self.update(dict(*args, **kwargs), allow_unknown_types=True)

    def __getitem__(self, key):
        """Get a (possibly nested) value from the struct.

        """
        if '.' in key:
            k, r = key.split('.', 1)
            # FIXME: this is inelegant.  better done with regexp?
            if len(k.split('[')) > 1:
                kl, i = k.split('[')
                i = int(i.strip(']'))
                return self.__dict__[kl][i][r]
            return self.__dict__[k][r]
        else:
            return self.__dict__[key]

    def get(self, key, default=None):
        """Get a (possibly nested) value from the struct, or default.

        """
        try:
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        if '.' in key:
            k, r = key.split('.', 1)
            self.__dict__[k][r] = value
        else:
            self.__dict__[key] = value

    def __delitem__(self, key):
        del self.__dict__[key]

    def setdefault(self, key, default):
        return self.__dict__.setdefault(key, default)

    def update(
            self, other,
            overwrite_atoms=False,
            clear_test=lambda v: False,
            allow_unknown_types=True,
    ):
        """Update Struct from other Struct or dict.
        This is *recursive* and will also update using lists, performing a
        deepcopy of the dict/list structure. It inspects the internal types to
        do this.

        None's are not inserted and are always overwritten.

        If a value of other returns true on clear_test(value) then that value is
        cleared in the updated self. override the argument
        clear_test=lambda v: v is None to clear null values.

        overwrite_atoms is an boolean argument which allows overwriting values
        with different types, like converting a float to a Struct. It defaults
        to False to disallow this often confusing behavior, but is sometimes
        necessary.

        allow_unknown_types allows updates and assignments of types other than
        Structs, Lists, floats, strings, and numpy arrays of elements. It
        defaults to True to be permissive.
        """
        kw = dict(
            overwrite_atoms=overwrite_atoms,
            clear_test=clear_test,
            allow_unknown_types=allow_unknown_types,
        )

        def update_element(self, k, other_v,):
            """
            type dispatch that assigns into self[k] based on the current type
            and the type of other_v
            """
            self_v = self[k]
            if clear_test(other_v):
                if isinstance(self, Mapping):
                    del self[k]
                else:
                    raise StructTypingError("clear_test deletions not allowed in sequences like lists")
            elif other_v is None:
                # don't update on None
                pass
            elif isinstance(other_v, VALUE_TYPES):
                # other is a value type, not a collection
                if isinstance(self_v, VALUE_TYPES):
                    self[k] = other_v
                elif isinstance(self_v, (Sequence, Mapping)):
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a value into a dict or list)")
                else:
                    if not allow_unknown_types:
                        raise StructTypingError("Unknown type assignment during recursive .update()")
                    else:
                        self[k] = other_v
            elif isinstance(other_v, Mapping):
                if isinstance(self_v, VALUE_TYPES):
                    if not overwrite_atoms:
                        raise StructTypingError("struct update is an incompatible storage type (e.g. updating a dict into a float)")
                    else:
                        self_v = self[k] = Struct()
                        self_v.update(other_v, **kw)
                elif isinstance(self_v, Sequence):
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a dict into a list)")
                elif isinstance(self_v, Mapping):
                    self[k].update(other_v, **kw)
                elif self_v is None:
                    self_v = self[k] = Struct()
                    self_v.update(other_v, **kw)
                else:
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a dict into a list)")
            elif isinstance(other_v, Sequence):
                # this check MUST come after VALUE_TYPES, or string is included

                # make mutable
                if not isinstance(self_v, MutableSequence):
                    self_v = list(self_v)

                if isinstance(self_v, VALUE_TYPES):
                    if not overwrite_atoms:
                        raise StructTypingError("struct update is an incompatible storage type (e.g. updating a dict into a string)")
                    else:
                        self_v = self[k] = other_v
                elif isinstance(self_v, Sequence):
                    # the string check MUST come before Sequence
                    list_update(self_v, other_v)
                elif isinstance(self_v, Mapping):
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a list into a dict)")
                elif self_v is None:
                    self_v = self[k] = other_v
                else:
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a value into a list)")
            else:
                # other is an unknown value type, not a collection
                if not allow_unknown_types:
                    raise StructTypingError("Unknown type assigned during recursive .update()")

                if isinstance(self_v, (Sequence, Mapping)):
                    raise StructTypingError("struct update is an incompatible storage type (e.g. updating a value into a dict or list)")
                else:
                    self[k] = other_v
            return

        def list_update(self_v, other_v,):
            """
            helper function for the recursive update
            """
            N_min = min(len(self_v), len(other_v))
            # make self as long as other, filled with None's so that assignment occurs
            self_v.extend([None] * (len(other_v) - N_min))
            for idx, sub_other_v in enumerate(other_v):
                update_element(self_v, idx, sub_other_v)
            return

        # actual code loop for the recursive update
        for k, other_v in other.items():
            if k in self:
                update_element(self, k, other_v)
            else:
                # k not in self, so just assign
                if clear_test(other_v):
                    pass
                elif isinstance(other_v, VALUE_TYPES):
                    # value type to directly assign
                    self[k] = other_v
                elif isinstance(other_v, Mapping):
                    self_v = self[k] = Struct()
                    # use update so that it is a deepcopy
                    self_v.update(other_v, **kw)
                elif isinstance(other_v, Sequence):
                    # MUST come after the value types check, or strings included
                    self_v = self[k] = []
                    list_update(self_v, other_v)
                else:
                    if not allow_unknown_types:
                        raise StructTypingError("Unknown type assigned during recursive .update()")
                    # value type to directly assign
                    self[k] = other_v

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __contains__(self, key):
        return key in self.__dict__

    def to_dict(self, array=False):
        """Return nested dictionary representation of Struct.

        If `array` is True, any lists encountered will be turned into
        numpy arrays, and lists of Structs will be turned into record
        arrays.  This is needed to convert to structure arrays in
        matlab.

        """
        d = {}
        for k, v in self.__dict__.items():
            if k[0] == '_':
                continue
            if isinstance(v, Struct):
                d[k] = v.to_dict(array=array)
            else:
                if isinstance(v, list):
                    try:
                        # this should fail if the elements of v are
                        # not Struct
                        # FIXME: need cleaner way to do this
                        v = [i.to_dict(array=array) for i in v]
                        if array:
                            v = dictlist2recarray(v)
                    except AttributeError:
                        if array:
                            v = np.array(v)
                # FIXME: there must be a better way to just match all
                # numeric scalar types
                elif isinstance(v, Number):
                    v = float(v)
                d[k] = v
        return d

    def to_yaml(self, path=None):
        """Return YAML representation of Struct.

        Write YAML to `path` if specified.

        """
        y = yaml.dump(self.to_dict(), default_flow_style=False)
        if path:
            with open(path, 'w') as f:
                f.write(y)
        else:
            return y

    def __str__(self):
        return '<GWINC Struct: {}>'.format(list(self.__dict__.keys()))

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.__dict__)

    def _repr_pretty_(self, s, cycle):
        """
        This is the pretty print extension function for IPython's pretty printer
        """
        if cycle:
            s.text('GWINC Struct(...)')
            return
        s.begin_group(8, 'Struct({')
        for idx, (k, v) in enumerate(self.items()):
            s.pretty(k)
            s.text(': ')
            s.pretty(v)
            if idx + 1 < len(self):
                s.text(',')
                s.breakable()
        s.end_group(8, '})')
        return

    def __iter__(self):
        return iter(self.__dict__)

    def walk(self):
        """Iterate over all leaves in the struct tree.

        """
        for k, v in self.__dict__.items():
            if k[0] == '_':
                continue
            if isinstance(v, (dict, Struct)):
                for sk, sv in v.walk():
                    yield k + '.' + sk, sv
            else:
                try:
                    for i, vv in enumerate(v):
                        if isinstance(vv, dict):
                            vv = Struct(vv)
                        for sk, sv in vv.walk():
                            yield '{}[{}].{}'.format(k, i, sk), sv
                except (AttributeError, TypeError):
                    yield k, v

    def hash(self, keys=None):
        """Hash of Struct.walk() data (recursive)

        """
        def filter_keys(kv):
            k, v = kv
            if keys:
                return k in keys
            else:
                return True

        def map_tuple(kv):
            k, v = kv
            if isinstance(v, list):
                return k, tuple(v)
            else:
                return k, v

        return hash(tuple(sorted(
            map(map_tuple, filter(filter_keys, self.walk()))
        )))

    def diff(self, other):
        """Return tuple of differences between target IFO.

        Returns list of (key, value, other_value) tuples.  Value is
        None if key not present.

        Note: yaml also supports putting None into dictionaries by not supplying
        a value. The None values returned here and the "missing value" None's
        from yaml are not distinguished in this diff
        """
        diffs = []

        # this is just a funky structure that will be guaranteed unique
        UNIQUE = (lambda x: None, )

        if isinstance(other, dict):
            other = Struct(other)
        for k, ov in other.walk():
            try:
                v = self.get(k, UNIQUE)
                if ov != v and ov is not v:
                    if v is UNIQUE:
                        diffs.append((k, None, ov))
                    else:
                        diffs.append((k, v, ov))
            except TypeError:
                # sometimes the deep keys go through unmappable objects
                # which TypeError if indexed
                diffs.append((k, None, ov))
        for k, v in self.walk():
            try:
                ov = other.get(k, UNIQUE)
                if ov is UNIQUE:
                    diffs.append((k, v, None))
            except TypeError:
                diffs.append((k, v, None))
        return diffs

    def __eq__(self, other):
        """True if structs have all equal values"""
        return not bool(self.diff(other))

    def to_txt(self, path=None, fmt='0.6e', delimiter=': ', end=''):
        """Return text represenation of Struct, one element per line.

        Struct keys use '.' to indicate hierarchy.  The `fmt` keyword
        controls the formatting of numeric values.  MATLAB code can be
        generated with the following parameters:

        >>> ifo.to_txt(delimiter=' = ', end=';')

        Write text to `path` if specified.

        """
        txt = io.StringIO()
        for k, v in sorted(self.walk()):
            if isinstance(v, (int, float, complex)):
                base = fmt
            elif isinstance(v, (list, np.ndarray)):
                if isinstance(v, list):
                    v = np.array(v)
                v = np.array2string(
                    v,
                    separator='',
                    max_line_width=np.Inf,
                    formatter={'all': lambda x: "{:0.6e} ".format(x)}
                )
                base = 's'
            else:
                base = 's'
            txt.write(u'{key}{delimiter}{value:{base}}{end}\n'.format(
                key=k, value=v, base=base,
                delimiter=delimiter,
                end=end,
            ))
        if path:
            with open(path, 'w') as f:
                f.write(txt.getvalue())
        else:
            return txt.getvalue()

    @classmethod
    def from_yaml(cls, y):
        """Create Struct from YAML string.

        """
        d = yaml.load(y, Loader=YAML_LOADER) or {}
        return cls(d)

    @classmethod
    def from_matstruct(cls, s):
        """Create Struct from scipy.io.matlab mat_struct object.

        """
        c = cls()
        try:
            s = s['ifo']
        except Exception:
            pass
        for k, v in s.__dict__.items():
            if k in ['_fieldnames']:
                # skip these fields
                pass
            elif type(v) is mat_struct:
                c.__dict__[k] = Struct.from_matstruct(v)
            else:
                # handle lists of Structs
                try:
                    c.__dict__[k] = list(map(Struct.from_matstruct, v))
                except Exception:
                    c.__dict__[k] = v
                    # try:
                    #     c.__dict__[k] = float(v)
                    # except:
                    #     c.__dict__[k] = v
        return c

    @classmethod
    def from_file(cls, path, _pass_inherit=False):
        """Load Struct from .yaml or MATLAB .mat file.

        Accepted file types are .yaml, .mat, or .m.

        For .m files, the file is expected to include either an object
        or function that corresponds to the basename of the file.  The
        MATLAB engine will be invoked to execute the .m code and
        extract the resultant IFO data.

        If `path` is a tuple, all elements will be joined ala
        os.path.join, with the first element resolved to it's absolute
        dirname.  This is useful for loading package-relative files
        with e.g.:

          Struct.from_file((__file__, 'myifo.yaml'))

        the _pass_inherit is a special key enabling the "+inherit" feature from
        load_budget. It will change in future versions as that functionality is
        moved here.
        """
        if type(path) == tuple:
            path = os.path.join(os.path.abspath(os.path.dirname(path[0])), *path[1:])
        base, ext = os.path.splitext(path)

        if ext == '.m':
            from .gwinc_matlab import Matlab
            matlab = Matlab()
            matlab.addpath(os.path.dirname(path))
            func_name = os.path.basename(base)
            matlab.eval("ifo = {};".format(func_name), nargout=0)
            ifo = matlab.extract('ifo')
            val = Struct.from_matstruct(ifo)

        else:
            with open(path, 'r') as f:
                if ext in ['.yaml', '.yml']:
                    val = cls.from_yaml(f)
                elif ext == '.mat':
                    s = loadmat(f, squeeze_me=True, struct_as_record=False)
                    val = cls.from_matstruct(s)
                else:
                    raise IOError("Unknown file type: {}".format(ext))

        # now include code it enable or disable "+inherit" at this stage of loading
        # this will modify in place if +inherit loading is active
        def recurse_value(v):
            if isinstance(v, VALUE_TYPES):
                pass
            elif isinstance(v, Sequence):
                recurse_sequence(v)
            elif isinstance(v, Mapping):
                recurse_mapping(v)
            else:
                pass

        def recurse_mapping(dct):
            for k, v in dct.items():
                if k == '+inherit':
                    raise RuntimeError('The +inherit key is not supported (yet) using from_file')
                recurse_value(v)

        def recurse_sequence(lst):
            for v in lst:
                recurse_value(v)

        # the bottom loop is special cased, as currently +inherit is only allowed at root level
        for k, v in val.items():
            if k == '+inherit':
                if _pass_inherit:
                    pass
                else:
                    raise RuntimeError('The +inherit key is not supported (yet) using from_file')
                continue
            recurse_value(v)
        return val


# these are the leaf types for struct that are most officially supported
# str must be included to distinguish it from Sequences like lists.
VALUE_TYPES = (str, Number, np.ndarray),


def dictlist2recarray(lst):
    def dtype(v):
        if isinstance(v, int):
            return float
        else:
            return type(v)
    # get dtypes from first element dict
    dtypes = [(k, dtype(v)) for k, v in lst[0].items()]
    values = [tuple(el.values()) for el in lst]
    out = np.array(values, dtype=dtypes)
    return out.view(np.recarray)


# HACK: fix loading Number in scientific notation
#
# https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number
#
# An apparent bug in python-yaml prevents it from recognizing
# scientific notation as a float.  The following is a modified version
# of the parser that recognize scientific notation appropriately.
YAML_LOADER = yaml.SafeLoader
YAML_LOADER.add_implicit_resolver(
    'tag:yaml.org,2002:float',
    re.compile('''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
    list('-+0123456789.'))


class StructTypingError(ValueError):
    """
    Exception class for specific errors related to typing during
    a recursive update to Struct.
    """
    pass


# add to the registry of mappings, as it is a useful way to write code to normalize between dicts and structs
Mapping.register(Struct)
